# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline manifest` commands:
    * snapshot
    * upload
    * diff
    * download
"""

from __future__ import annotations

from configparser import ConfigParser
import dataclasses
import os
import sys
from typing import List, Optional
import boto3
import click

from deadline.client import api
from deadline.client.config import config_file
from deadline.job_attachments._diff import pretty_print_cli
from deadline.job_attachments._utils import (
    WINDOWS_MAX_PATH_LENGTH,
    _is_windows_long_path_registry_enabled,
)
from deadline.job_attachments.api.manifest import (
    _glob_files,
    _manifest_diff,
    _manifest_download,
    _manifest_snapshot,
    _manifest_upload,
)
from deadline.job_attachments.models import (
    S3_MANIFEST_FOLDER_NAME,
    JobAttachmentS3Settings,
    ManifestDiff,
    AssetType,
)

from ...exceptions import NonValidInputError
from .._common import _apply_cli_options_to_config, _handle_error
from .._main import deadline as main
from .click_logger import ClickLogger


@main.group(name="manifest")
@_handle_error
def cli_manifest():
    """
    BETA - Commands to work with [Deadline Cloud job attachments].

    [Deadline Cloud job attachments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/storage-job-attachments.html
    """


@cli_manifest.command(name="snapshot")
@click.option("--root", required=True, help="The root directory to snapshot. ")
@click.option(
    "-d",
    "--destination",
    default=None,
    help="Destination directory where manifest is created. Defaults to the manifest root directory.",
)
@click.option(
    "-n",
    "--name",
    default=None,
    help="Name of the manifest. A timestamp is added YYYY-MM-DD-HH-MM-SS for versioning.",
)
@click.option(
    "-i",
    "--include",
    default=None,
    help="Glob syntax of files and directories to include in the manifest. Can be provided multiple times.",
    multiple=True,
)
@click.option(
    "-e",
    "--exclude",
    default=None,
    help="Glob syntax of files and directories to exclude in the manifest. Can be provided multiple times.",
    multiple=True,
)
@click.option(
    "-ie",
    "--include-exclude-config",
    default=None,
    help="Include and exclude config of files and directories to include and exclude. Can be a json file or json string.",
    multiple=True,
)
@click.option(
    "--force-rehash",
    default=False,
    is_flag=True,
    help="Rehash all files to compare using file hashes.",
)
@click.option("--diff", default=None, help="File Path to Asset Manifest to diff against.")
@click.option("--json", default=None, is_flag=True, help="Output is printed as JSON for scripting.")
@_handle_error
def manifest_snapshot(
    root: str,
    destination: str,
    name: str,
    include: List[str],
    exclude: List[str],
    include_exclude_config: str,
    diff: str,
    force_rehash: bool,
    json: bool,
    **args,
):
    """
    BETA - Generates a snapshot of files in a directory root as a [job attachments] Manifest.

    [job attachments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/storage-job-attachments.html
    """
    logger: ClickLogger = ClickLogger(is_json=json)
    if not os.path.isdir(root):
        raise NonValidInputError(f"Specified root directory {root} does not exist.")

    if destination and not os.path.isdir(destination):
        raise NonValidInputError(f"Specified destination directory {destination} does not exist.")
    elif destination is None:
        destination = root
        logger.echo(f"Manifest creation path defaulted to {root} \n")

    manifest_out = _manifest_snapshot(
        root=root,
        destination=destination,
        name=name,
        include=include,
        exclude=exclude,
        include_exclude_config=include_exclude_config,
        diff=diff,
        force_rehash=force_rehash,
        print_function_callback=logger.echo,
    )
    if manifest_out:
        if (
            sys.platform == "win32"
            and len(manifest_out.manifest) >= WINDOWS_MAX_PATH_LENGTH
            and not _is_windows_long_path_registry_enabled()
        ):
            long_manifest_path_warning = f"""WARNING: Manifest file path {manifest_out.manifest} exceeds Windows path length limit. This may cause unexpected issues.
For details and a fix using the registry, see: https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation"""
            logger.echo(
                click.style(
                    long_manifest_path_warning,
                    fg="yellow",
                )
            )
            logger.json(
                dict(dataclasses.asdict(manifest_out), **{"warning": long_manifest_path_warning})
            )
        else:
            logger.json(dataclasses.asdict(manifest_out))


@cli_manifest.command(name="diff")
@click.option("--root", help="The root directory to compare changes to.")
@click.option(
    "--manifest",
    required=True,
    help="The path to manifest file to diff against.",
)
@click.option(
    "-i",
    "--include",
    default=None,
    help="Glob syntax of files and directories to include in the manifest. Can be provided multiple times.",
    multiple=True,
)
@click.option(
    "-e",
    "--exclude",
    default=None,
    help="Glob syntax of files and directories to exclude in the manifest. Can be provided multiple times.",
    multiple=True,
)
@click.option(
    "-ie",
    "--include-exclude-config",
    default=None,
    help="Include and exclude config of files and directories to include and exclude. Can be a json file or json string.",
)
@click.option(
    "--force-rehash",
    default=False,
    is_flag=True,
    help="Rehash all files to compare using file hashes.",
)
@click.option("--json", default=None, is_flag=True, help="Output is printed as JSON for scripting.")
@_handle_error
def manifest_diff(
    root: str,
    manifest: str,
    include: List[str],
    exclude: List[str],
    include_exclude_config: str,
    force_rehash: bool,
    json: bool,
    **args,
):
    """
    BETA - Compute the file difference of a root directory against an existing
    [job attachments] manifest for new, modified or deleted files.

    [job attachments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/storage-job-attachments.html
    """
    logger: ClickLogger = ClickLogger(is_json=json)
    if not os.path.isfile(manifest):
        raise NonValidInputError(f"Specified manifest file {manifest} does not exist. ")

    if not os.path.isdir(root):
        raise NonValidInputError(f"Specified root directory {root} does not exist. ")

    # Perform the diff.
    differences: ManifestDiff = _manifest_diff(
        manifest=manifest,
        root=root,
        include=include,
        exclude=exclude,
        include_exclude_config=include_exclude_config,
        force_rehash=force_rehash,
        print_function_callback=logger.echo,
    )

    # Print results to console.
    if json:
        logger.json(dataclasses.asdict(differences), indent=4)
    else:
        logger.echo(f"Manifest Diff of root directory: {root}")
        all_files = _glob_files(
            root=root,
            include=include,
            exclude=exclude,
            include_exclude_config=include_exclude_config,
        )
        pretty_print_cli(root=root, all_files=all_files, manifest_diff=differences)


@cli_manifest.command(name="download")
@click.argument("download_dir")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--job-id", required=True, help="The AWS Deadline Cloud Job to get. ")
@click.option("--step-id", help="The AWS Deadline Cloud Step to get. ")
@click.option("--farm-id", help="The AWS Deadline Cloud Farm to use. ")
@click.option("--queue-id", help="The AWS Deadline Cloud Queue to use. ")
@click.option(
    "--asset-type",
    default=AssetType.ALL.value,
    help="Which asset type to download:\n"
    "INPUT means download only input asset files for given job/step.\n"
    "OUTPUT means download only output asset files for given job/step.\n"
    "ALL (default) means download all input & output asset files for given job/step.\n",
    type=click.Choice(
        [e.value for e in AssetType],
        case_sensitive=False,
    ),
)
@click.option(
    "--json", default=None, is_flag=True, help="Output is printed as JSON for scripting. "
)
@_handle_error
def manifest_download(
    download_dir: str,
    job_id: str,
    step_id: str,
    asset_type: str,
    json: bool,
    **args,
):
    """
    BETA - Download [job attachments] Manifests for a Job, or Step including dependencies.

    [job attachments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/storage-job-attachments.html
    """
    logger: ClickLogger = ClickLogger(is_json=json)
    if not os.path.isdir(download_dir):
        raise NonValidInputError(f"Specified destination directory {download_dir} does not exist. ")

    # setup config
    config: Optional[ConfigParser] = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id"}, **args
    )
    queue_id: str = config_file.get_setting("defaults.queue_id", config=config)
    farm_id: str = config_file.get_setting("defaults.farm_id", config=config)

    boto3_session: boto3.Session = api.get_boto3_session(config=config)

    output = _manifest_download(
        download_dir=download_dir,
        farm_id=farm_id,
        queue_id=queue_id,
        job_id=job_id,
        step_id=step_id,
        asset_type=AssetType(asset_type),
        boto3_session=boto3_session,
        print_function_callback=logger.echo,
    )
    logger.json(dataclasses.asdict(output))


@cli_manifest.command(name="upload")
@click.argument("manifest_file")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--s3-cas-uri", help="The URI to the Content Addressable Storage S3 bucket and root.")
@click.option(
    "--s3-manifest-prefix", help="Prefix subpath in the manifest folder to upload the manifest."
)
@click.option(
    "--farm-id", help="The AWS Deadline Cloud Farm to use. Alternative to using --s3-cas-uri."
)
@click.option(
    "--queue-id", help="The AWS Deadline Cloud Queue to use. Alternative to using --s3-cas-uri."
)
@click.option("--json", default=None, is_flag=True, help="Output is printed as JSON for scripting.")
@_handle_error
def manifest_upload(
    manifest_file: str,
    s3_cas_uri: str,
    s3_manifest_prefix: str,
    json: bool,
    **args,
):
    """
    BETA - Uploads a [job attachments] manifest file to a Content Addressable Storage's Manifest store.
    If calling via --s3-cas-path, it is recommended to use with --profile for a specific AWS profile
    with CAS S3 bucket access. Check exit code for success or failure.

    [job attachments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/storage-job-attachments.html
    """
    # Input checking.
    if not manifest_file or not os.path.isfile(manifest_file):
        raise NonValidInputError(f"Specified manifest {manifest_file} does not exist. ")

    # Where will we upload the manifest to?
    required: set[str] = set()
    if not s3_cas_uri:
        required = {"farm_id", "queue_id"}

    config: Optional[ConfigParser] = _apply_cli_options_to_config(required_options=required, **args)

    # Logger
    logger: ClickLogger = ClickLogger(is_json=json)

    bucket_name: str = ""
    session: boto3.Session = api.get_boto3_session(config=config)
    if not s3_cas_uri:
        farm_id = config_file.get_setting("defaults.farm_id", config=config)
        queue_id = config_file.get_setting("defaults.queue_id", config=config)

        deadline = api.get_boto3_client("deadline", config=config)
        queue = deadline.get_queue(
            farmId=farm_id,
            queueId=queue_id,
        )
        queue_ja_settings: JobAttachmentS3Settings = JobAttachmentS3Settings(
            **queue["jobAttachmentSettings"]
        )
        bucket_name = queue_ja_settings.s3BucketName
        cas_path = queue_ja_settings.rootPrefix

        # IF we supplied a farm and queue, use the queue credentials.
        session = api.get_queue_user_boto3_session(
            deadline=deadline,
            config=config,
            farm_id=farm_id,
            queue_id=queue_id,
            queue_display_name=queue["displayName"],
        )

    else:
        # Self supplied cas path.
        uri_ja_settings: JobAttachmentS3Settings = JobAttachmentS3Settings.from_s3_root_uri(
            s3_cas_uri
        )
        bucket_name = uri_ja_settings.s3BucketName
        cas_path = uri_ja_settings.rootPrefix

    logger.echo(
        f"Uploading Manifest to {bucket_name} {cas_path} {S3_MANIFEST_FOLDER_NAME}, prefix: {s3_manifest_prefix}"
    )
    _manifest_upload(
        manifest_file=manifest_file,
        s3_bucket_name=bucket_name,
        s3_cas_prefix=cas_path,
        s3_key_prefix=s3_manifest_prefix,
        boto_session=session,
        print_function_callback=logger.echo,
    )
    logger.echo("Uploading successful!")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides a modal dialog box for displaying submitter and environment information about the open submitter.
"""

from __future__ import annotations

import logging
import yaml

from dataclasses import asdict
from typing import List, Dict
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QDialog,
    QDialogButtonBox,
    QLabel,
    QMessageBox,
    QPushButton,
    QTextEdit,
    QVBoxLayout,
)
from qtpy.QtCore import Qt  # pylint: disable=import-error

from ...dataclasses import SubmitterInfo
from ..dataclasses._environment_info import _EnvironmentInfo
from .._utils import tr

logger = logging.getLogger(__name__)


# Hard-coded documentation links
_DOCUMENTATION_LINKS: List[Dict[str, str]] = [
    {"name": "User Guide", "url": "https://aws-deadline.github.io"},
]


class _HelpDialog(QDialog):
    """
    Modal dialog that displays submitter and environment information about the deadline-cloud
    library and the environment it's running in.

    The dialog displays all of the following if available:
    - Submitter name
    - Documentation link
    - Deadline Cloud library version
    - Submitter package name and version
    - Host application name and version
    - Additional application provided information
    - Python version
    - Operating system version
    - Qt version
    """

    def __init__(self, submitter_info: SubmitterInfo, parent=None):
        """
        Initialize the Help dialog.

        Args:
            submitter_info: SubmitterInfo object containing submitter details
            parent: Parent widget (optional)
        """
        super().__init__(parent=parent)

        self.submitter_info = submitter_info
        self.environment_info = _EnvironmentInfo.collect()

        self.setWindowTitle(
            f"{tr('About')} Deadline Cloud {submitter_info.submitter_name} {tr('Submitter')}"
        )
        self.setModal(True)
        self.setMinimumWidth(500)

        self._build_ui()

    def _build_ui(self):
        """Build the dialog UI layout."""
        layout = QVBoxLayout()

        # Add documentation links section (hard-coded)
        for link_info in _DOCUMENTATION_LINKS:
            name = link_info["name"]
            url = link_info["url"]
            doc_label = QLabel()
            doc_label.setText(f"<a href='{url}'>{name}</a>")
            doc_label.setToolTip(url)
            doc_label.setOpenExternalLinks(True)
            doc_label.setTextInteractionFlags(Qt.TextBrowserInteraction)
            doc_label.setWordWrap(True)
            layout.addWidget(doc_label)

        self.version_text = QTextEdit()
        self.version_text.setReadOnly(True)
        self.version_text.setPlainText(self._format_version_info())
        self.version_text.setMinimumHeight(200)
        layout.addWidget(self.version_text)

        button_box = QDialogButtonBox()

        self.copy_button = QPushButton(tr("Copy"))
        self.copy_button.clicked.connect(self._on_copy_clicked)
        button_box.addButton(self.copy_button, QDialogButtonBox.ActionRole)

        close_button = button_box.addButton(QDialogButtonBox.Close)
        close_button.setText(tr("Close"))
        close_button.clicked.connect(self.accept)

        layout.addWidget(button_box)

        self.setLayout(layout)

    @staticmethod
    def _make_keys_human_readable(data):
        """
        Recursively replace underscores with spaces in dictionary keys.

        Args:
            data: The data structure to process (dict, list, or scalar value)

        Returns:
            The data structure with all dictionary keys made human-readable
        """
        if isinstance(data, dict):
            return {
                k.replace("_", " "): _HelpDialog._make_keys_human_readable(v)
                for k, v in data.items()
            }
        elif isinstance(data, list):
            return [_HelpDialog._make_keys_human_readable(item) for item in data]
        else:
            return data

    def _format_version_info(self) -> str:
        """
        Combine submitter and version information for display.

        Returns:
            Formatted yaml string with submitter and version information
        """

        env_data = asdict(self.environment_info)
        submitter_data = asdict(self.submitter_info)

        # Filter out None values from submitter data
        submitter_dict = {k: v for k, v in submitter_data.items() if v is not None}

        combined_info = {**submitter_dict, **env_data}

        # Replace underscores with spaces in keys for user-friendly display (recursively)
        friendly_info = self._make_keys_human_readable(combined_info)

        yaml_str = yaml.dump(friendly_info, indent=4, sort_keys=False)
        return yaml_str

    def _format_for_copy(self) -> str:
        """
        Format all information for copy/pasting with header.

        Returns:
            Formatted string suitable for copying and pasting
        """
        header = "AWS Deadline Cloud Submitter Information"
        lines = [header, "=" * len(header)]
        lines.append(self._format_version_info())
        return "\n".join(lines)

    def _on_copy_clicked(self):
        """Handle the Copy button click."""
        try:
            clipboard = QApplication.clipboard()
            if clipboard:
                formatted_text = self._format_for_copy()
                clipboard.setText(formatted_text)
                logger.info("Submitter information copied to clipboard")
            else:
                logger.warning("Clipboard is not available")
                QMessageBox.warning(
                    self,
                    "Copy Failed",
                    "Unable to access the system clipboard.",
                )
        except Exception as e:
            logger.error(f"Failed to copy version info: {e}")
            QMessageBox.warning(
                self,
                "Copy Failed",
                f"Failed to copy version information to clipboard: {str(e)}",
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import json
import os
from typing import Any
from ._yaml import deadline_yaml_dump


def save_yaml_or_json_to_file(
    bundle_dir: str,
    filename: str,
    file_type: str,
    data: Any,
) -> None:
    """
    Saves data as either a JSON or YAML file depending on the file_type provided. Useful for saving
    job bundle data files which can be in either format. file_type should be either "YAML" or "JSON".
    """
    with open(
        os.path.join(bundle_dir, f"{filename}.{file_type.lower()}"), "w", encoding="utf8"
    ) as f:
        if file_type == "YAML":
            deadline_yaml_dump(data, f)
        elif file_type == "JSON":
            json.dump(data, f, indent=2)
        else:
            raise RuntimeError(f"Unexpected file type '{file_type}' in job bundle:\n{bundle_dir}")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = ["ControlType"]
import enum


class ControlType(enum.Enum):
    """
    Defines all the valid GUI control types that are supported in an Open Job Description
    job template parameter's "userInterface" property.
    """

    LINE_EDIT = enum.auto()
    MULTILINE_EDIT = enum.auto()
    # SPIN_BOX for INT type
    INT_SPIN_BOX = enum.auto()
    # SPIN_BOX for FLOAT type
    FLOAT_SPIN_BOX = enum.auto()
    DROPDOWN_LIST = enum.auto()
    CHOOSE_INPUT_FILE = enum.auto()
    CHOOSE_OUTPUT_FILE = enum.auto()
    CHOOSE_DIRECTORY = enum.auto()
    CHECK_BOX = enum.auto()
    HIDDEN = enum.auto()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Exceptions that the Deadline Job Attachments library can raise.
"""

from typing import Optional


COMMON_ERROR_GUIDANCE_FOR_S3 = {
    408: "Request timeout. Please consider retrying later, or ensure your network connection is stable.",
    500: "Internal server error. It might be an issue on AWS's side; please consider retrying later or contacting AWS support.",
    503: "Service unavailable. AWS S3 might be down or experiencing high traffic. Please consider retrying after some time.",
}


class AssetSyncError(Exception):
    """
    Exception for errors related to synching files to/from S3.
    """


class JobAttachmentsError(Exception):
    """
    Exception for errors related to the Deadline Service.
    """


class JobAttachmentsS3ClientError(AssetSyncError):
    """
    Exception for errors related to the S3 client.
    """

    def __init__(
        self,
        action,
        status_code,
        bucket_name: str,
        key_or_prefix: str,
        message: Optional[str] = None,
    ) -> None:
        self.action = action
        self.status_code = status_code
        self.bucket_name = bucket_name
        self.key_or_prefix = key_or_prefix

        message_parts = [
            f"Error {action} in bucket '{bucket_name}', Target key or prefix: '{key_or_prefix}'",
            f"HTTP Status Code: {status_code}",
        ]
        if message:
            message_parts.append(message)

        super().__init__(", ".join(message_parts))


class JobAttachmentS3BotoCoreError(AssetSyncError):
    """
    Exception to wrap any botocore.exceptions.BotoCoreError.
    """

    def __init__(self, action: str, error_details: str) -> None:
        self.action = action
        message = (
            f"An issue occurred with AWS service request while {action}: "
            f"{error_details}\n"
            "This could be due to temporary issues with AWS, internet connection, or your AWS credentials. "
            "Please verify your credentials and network connection. If the problem persists, try again later"
            " or contact support for further assistance."
        )
        super().__init__(message)


class MissingJobAttachmentSettingsError(JobAttachmentsError):
    """
    Exception raised when attempting to use Job Attachments but the settings are not set in Queue.
    """


class ManifestCreationException(Exception):
    """
    Exception for errors related to Creating Manifests.
    """


class MissingS3BucketError(JobAttachmentsError):
    """
    Exception raised when attempting to use Job Attachments but the S3 bucket is not set in Queue.
    """


class MissingS3RootPrefixError(JobAttachmentsError):
    """
    Exception raised when attempting to use Job Attachments but the S3 root prefix is not set in Queue.
    """


class MalformedAttachmentSettingError(JobAttachmentsError):
    """
    Exception raised when encountering error parsing input attachment settings.
    """


class AssetOutsideOfRootError(JobAttachmentsError):
    """
    Exception for errors related to assets being outside of the asset root.
    """


class MisconfiguredInputsError(JobAttachmentsError):
    """
    Exception for errors related to missing input directories, empty input directories,
    missing input files, and input directories classified as files
    """


class ManifestDecodeValidationError(JobAttachmentsError):
    """
    Exception for errors related to asset manifest decoding.
    """


class MissingManifestError(JobAttachmentsError):
    """
    Exception for when trying to retrieve asset manifests that don't exist.
    """


class MissingAssetRootError(JobAttachmentsError):
    """
    Exception for when trying to retrieve asset root from metadata (in S3) that doesn't exist.
    """


class AssetSyncCancelledError(JobAttachmentsError):
    """
    Exception thrown when an operation (synching files to/from S3) has been cancelled.
    """

    def __init__(self, message, summary_statistics=None):
        super().__init__(message)
        self.summary_statistics = summary_statistics


class PathOutsideDirectoryError(JobAttachmentsError):
    """
    Exception thrown in the _ensure_paths_within_directory function to signal that a given
    file path, especially ones that may contain "..", does not reside in the specified root path.
    """


class VFSExecutableMissingError(JobAttachmentsError):
    """
    Exception for when trying to retrieve VFS executable path doesn't exist.
    """


class VFSLaunchScriptMissingError(JobAttachmentsError):
    """
    Exception for when trying to retrieve VFS launch script path doesn't exist.
    """


class VFSFailedToMountError(JobAttachmentsError):
    """
    Exception for when trying to mount VFS at a given path.
    """


class VFSOSUserNotSetError(JobAttachmentsError):
    """
    Exception attempting to use the vfs without an os user
    """


class UnsupportedHashingAlgorithmError(JobAttachmentsError):
    """
    Exception for when an unsupported hashing algorithm is provided.
    """


class VFSRunPathNotSetError(JobAttachmentsError):
    """
    Exception for when the run path hasn't been set for the vfs
    """



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline config` commands.
"""

import click
import json
import textwrap

from ...config import config_file
from .._common import _handle_error
from .._main import deadline as main


@main.group(name="config")
@_handle_error
def cli_config():
    """
    Commands to show and update Deadline's workstation configuration.

    Deadline's workstation configuration is organized in a shallow hierarchy.
    Most of the settings are set to depend on the selected AWS profile, so that
    you can store different configurations like farm and queue for different profiles
    and switch between them by selecting the profile.
    """


@cli_config.command(name="show")
@click.option(
    "--output",
    type=click.Choice(["verbose", "json"], case_sensitive=False),
    default="verbose",
    help="Output format of the command",
)
@_handle_error
def config_show(output):
    """
    Show all workstation configuration settings and current values.
    """
    settings_json = {}
    if output == "verbose":
        click.echo(
            f"AWS Deadline Cloud configuration file:\n   {config_file.get_config_file_path()}"
        )
        click.echo()
        for setting_name in config_file.SETTINGS.keys():
            setting_value = config_file.get_setting(setting_name)
            setting_default = config_file.get_setting_default(setting_name)

            # Wrap and indent the descriptions to 80 characters because they may be multiline.
            setting_description: str = config_file.SETTINGS[setting_name].get("description", "")
            setting_description = "\n".join(
                f"   {line}" for line in textwrap.wrap(setting_description, width=77)
            )
            click.echo(
                f"{setting_name}: {setting_value} {'(default)' if setting_value == setting_default else ''}"
            )
            click.echo(setting_description)
            click.echo()
    else:
        settings_json["settings.config_file_path"] = str(config_file.get_config_file_path())
        for setting_name in config_file.SETTINGS.keys():
            setting_value = config_file.get_setting(setting_name)
            settings_json[setting_name] = setting_value

        click.echo(json.dumps(settings_json))


@cli_config.command(name="gui")
@click.option(
    "--install-gui",
    is_flag=True,
    help="Installs GUI dependencies if they are not installed already",
)
@_handle_error
def config_gui(install_gui: bool):
    """
    Open the workstation configuration settings GUI to view or edit setting values.
    """
    from ...ui import gui_context_for_cli

    with gui_context_for_cli(automatically_install_dependencies=install_gui):
        from ...ui.dialogs.deadline_config_dialog import DeadlineConfigDialog

        DeadlineConfigDialog.configure_settings()


@cli_config.command(name="set")
@click.argument("setting_name")
@click.argument("value")
@_handle_error
def config_set(setting_name, value):
    """
    Sets a workstation configuration setting.

    \b
    Example:
        `deadline config set defaults.farm_id <farm-id>`
    """
    config_file.set_setting(setting_name, value)


@cli_config.command(name="clear")
@click.argument("setting_name")
@_handle_error
def config_clear(setting_name):
    """
    Clears a workstation configuration setting to restore its default value.

    \b
    Example:
        `deadline config clear defaults.farm_id`
    """
    config_file.clear_setting(setting_name)


@cli_config.command(name="get")
@click.argument("setting_name")
@_handle_error
def config_get(setting_name):
    """
    Prints the value of a workstation configuration setting. You can use this command
    to call the [AWS CLI][aws-cli] with settings from the Deadline workstation configuration.

    [aws-cli]: https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-welcome.html

    \b
    Example:
        `aws deadline get-queue --profile $(deadline config get defaults.aws_profile_name) --farm-id $(deadline config get defaults.farm_id) --queue-id $(deadline config get defaults.queue_id)`
    """
    click.echo(config_file.get_setting(setting_name))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Data classes for AWS objects.
"""

from __future__ import annotations

import json
import sys
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional, Set
from urllib.parse import urlparse

from deadline.job_attachments.asset_manifests import HashAlgorithm, hash_data
from deadline.job_attachments.asset_manifests.base_manifest import (
    BaseAssetManifest,
    BaseManifestPath,
)
from deadline.job_attachments.exceptions import (
    MissingS3RootPrefixError,
    MalformedAttachmentSettingError,
)

from ._utils import (
    _generate_random_guid,
    _join_s3_paths,
    _float_to_iso_datetime_string,
)

S3_DATA_FOLDER_NAME = "Data"
S3_MANIFEST_FOLDER_NAME = "Manifests"
S3_INPUT_MANIFEST_FOLDER_NAME = "Inputs"


@dataclass
class AssetRootManifest:
    """Represents asset manifest and a list of output files grouped under the same root"""

    file_system_location_name: Optional[str] = None
    root_path: str = ""
    asset_manifest: Optional[BaseAssetManifest] = None
    outputs: List[Path] = field(default_factory=list)


@dataclass
class AssetRootGroup:
    """Represents lists of input files, output files and path references grouped under the same root"""

    file_system_location_name: Optional[str] = None
    root_path: str = ""
    inputs: Set[Path] = field(default_factory=set)
    outputs: Set[Path] = field(default_factory=set)
    references: Set[Path] = field(default_factory=set)


@dataclass
class AssetUploadGroup:
    """Represents all of the information needed to prepare to upload assets"""

    asset_groups: List[AssetRootGroup] = field(default_factory=list)
    known_asset_paths: List[Path] = field(default_factory=list)
    """List of paths that should not generate warnings"""
    total_input_files: int = 0
    total_input_bytes: int = 0


@dataclass
class ManifestPathGroup:
    """
    Represents paths combined from multiple manifests under the same root path, organized by hash algorithm.
    """

    total_bytes: int = 0
    files_by_hash_alg: dict[HashAlgorithm, List[BaseManifestPath]] = field(default_factory=dict)

    def add_manifest_to_group(self, manifest: BaseAssetManifest) -> None:
        if manifest.hashAlg not in self.files_by_hash_alg:
            self.files_by_hash_alg[manifest.hashAlg] = manifest.paths
        else:
            self.files_by_hash_alg[manifest.hashAlg].extend(manifest.paths)
        self.total_bytes += manifest.totalSize  # type: ignore[attr-defined]

    def combine_with_group(self, group: ManifestPathGroup) -> None:
        """Adds the content of the given ManifestPathGroup to this ManifestPathGroup"""
        for hash_alg, paths in group.files_by_hash_alg.items():
            if hash_alg not in self.files_by_hash_alg:
                self.files_by_hash_alg[hash_alg] = paths
            else:
                self.files_by_hash_alg[hash_alg].extend(paths)
            self.total_bytes += group.total_bytes

    def get_all_paths(self) -> list[str]:
        """
        Get all paths in this group, regardless of hashing algorithm.
        Note that this may include duplicates if the same path exists for multiple hashing algorithms.

        Returns a sorted list of paths represented as strings.
        """
        path_list: List[str] = []
        for paths in self.files_by_hash_alg.values():
            path_list.extend([path.path for path in paths])
        return sorted(path_list)


@dataclass
class OutputFile:
    """Files for output"""

    # File size in Bytes
    file_size: int
    file_hash: str
    rel_path: str
    full_path: str
    s3_key: str
    # If the file already exists in the CAS
    in_s3: bool
    # The base directory path against which file paths are containment-checked
    base_dir: Optional[str]


class StorageProfileOperatingSystemFamily(str, Enum):
    """Case-insensitive enum for the storage profile operating system family type."""

    WINDOWS = "windows"
    LINUX = "linux"
    MACOS = "macos"

    @classmethod
    def _missing_(cls, value):
        value = value.lower()
        for member in cls:
            if member == value:
                return member
        return None

    @classmethod
    def get_host_os_family(cls) -> StorageProfileOperatingSystemFamily:
        """Get the current path format."""
        if sys.platform.startswith("win"):
            return cls.WINDOWS
        if sys.platform.startswith("darwin"):
            return cls.MACOS
        if sys.platform.startswith("linux"):
            return cls.LINUX
        else:
            raise NotImplementedError(f"Operating system {sys.platform} is not supported.")


class AssetType(str, Enum):
    INPUT = "input"
    OUTPUT = "output"
    ALL = "all"


class PathFormat(str, Enum):
    WINDOWS = "windows"
    POSIX = "posix"

    @classmethod
    def get_host_path_format(cls) -> PathFormat:
        """Get the current path format."""
        if sys.platform.startswith("win"):
            return cls.WINDOWS
        if sys.platform.startswith("darwin") or sys.platform.startswith("linux"):
            return cls.POSIX
        else:
            raise NotImplementedError(f"Operating system {sys.platform} is not supported.")

    @classmethod
    def get_host_path_format_string(cls) -> str:
        """Get a string of the current path format."""
        return cls.get_host_path_format().value


# Behavior to adopt when loading job assets
class JobAttachmentsFileSystem(str, Enum):
    # Load all assets at before execution of the job code
    COPIED = "COPIED"
    # Start job execution immediately and load assets as needed
    VIRTUAL = "VIRTUAL"


@dataclass
class ManifestProperties:
    """The assets for a Step under an asset root"""

    # The path that assets were relative to on submitting machine
    rootPath: str
    # Used for path mapping.
    rootPathFormat: PathFormat
    # If submitting machine has a 'Local' Storage Profile and files are relative
    # to any of its 'Asset Roots', the Asset Root Path will be used below.
    # Otherwise, the dynamic Job Attachments root path will be used and this will be empty.
    fileSystemLocationName: Optional[str] = field(default=None)  # type: ignore
    # An S3 (object) key that points to a file manifest location.
    # Optional as we may not need inputs if everything is embedded in the Job Template.
    inputManifestPath: Optional[str] = field(default=None)  # type: ignore
    # The hash of the manifest, for data provenance
    inputManifestHash: Optional[str] = field(default=None)  # type: ignore
    # The expected output directories to search for outputs. Relative to the rootPath.
    outputRelativeDirectories: Optional[List[str]] = field(default=None)

    def to_dict(self) -> dict[str, Any]:
        result: dict[str, Any] = {"rootPath": self.rootPath}
        if self.fileSystemLocationName:
            result["fileSystemLocationName"] = self.fileSystemLocationName
        result["rootPathFormat"] = self.rootPathFormat.value
        if self.inputManifestPath:
            result["inputManifestPath"] = self.inputManifestPath
        if self.inputManifestHash:
            result["inputManifestHash"] = self.inputManifestHash
        if self.outputRelativeDirectories:
            result["outputRelativeDirectories"] = self.outputRelativeDirectories
        return result

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "ManifestProperties":
        """Create ManifestProperties from a dictionary."""
        return cls(
            rootPath=data["rootPath"],
            rootPathFormat=PathFormat(data["rootPathFormat"]),
            fileSystemLocationName=data.get("fileSystemLocationName"),
            inputManifestPath=data.get("inputManifestPath"),
            inputManifestHash=data.get("inputManifestHash"),
            outputRelativeDirectories=data.get("outputRelativeDirectories"),
        )

    def as_output_metadata(self) -> dict[str, dict[str, str]]:
        """
        Generate S3 metadata for output manifest uploads.

        Creates metadata dictionary containing asset root path and optional file system location.
        Handles non-ASCII characters in paths by JSON-encoding them with ASCII-safe format.

        Returns:
            dict[str, str]: S3 metadata dictionary with 'Metadata' key containing:
                - 'asset-root': ASCII-compatible root path, or JSON-encoded root path for non-ASCII paths
                - 'asset-root-json': JSON-encoded root path for non-ASCII paths
                - 'file-system-location-name': Optional file system location name
        """
        metadata: dict[str, str] = {}
        try:
            # Set 'asset-root' metadata as the path if the path is ASCII
            self.rootPath.encode(encoding="ascii")
            metadata["asset-root"] = self.rootPath
        except UnicodeEncodeError:
            # S3 metadata must be ASCII
            # Add both 'asset-root' and 'asset-root-json' metadata encoded to ASCII as a JSON string
            # Populate both fileds for backward compatibility
            json_root_path = json.dumps(self.rootPath, ensure_ascii=True)
            metadata["asset-root-json"] = json_root_path
            metadata["asset-root"] = json_root_path
        if self.fileSystemLocationName:
            metadata["file-system-location-name"] = self.fileSystemLocationName

        return {"Metadata": metadata}


@dataclass
class Attachments:
    """An object that holds the job attachments for a Job"""

    # The list of required assets per asset root
    manifests: List[ManifestProperties] = field(default_factory=list)
    # Method to use when loading assets required for a job
    fileSystem: str = JobAttachmentsFileSystem.COPIED.value

    def to_dict(self) -> dict[str, Any]:
        return {
            "manifests": [manifest.to_dict() for manifest in self.manifests],
            "fileSystem": self.fileSystem,
        }


@dataclass
class JobAttachmentS3Settings:
    """S3-specific Job Attachment settings, configured at the Queue level."""

    # The S3 bucket all attachments are stored in. (required)
    s3BucketName: str  # pylint: disable=invalid-name
    # The S3 bucket prefix all files are stored relative to. (required)
    rootPrefix: str  # pylint: disable=invalid-name

    @staticmethod
    def from_root_path(root_path: str) -> JobAttachmentS3Settings:
        path_split: list = root_path.split("/")

        if len(path_split) < 2:
            raise MalformedAttachmentSettingError(
                "Invalid root path format, should be s3BucketName/rootPrefix."
            )

        return JobAttachmentS3Settings(path_split[0], "/".join(path_split[1:]))

    @staticmethod
    def from_s3_root_uri(uri: str) -> JobAttachmentS3Settings:
        res = urlparse(uri)

        if not res.netloc or not res.path[1:] or res.scheme != "s3":
            raise MalformedAttachmentSettingError(
                "Invalid root uri format, should be s3://s3BucketName/rootPrefix."
            )

        return JobAttachmentS3Settings(res.netloc, res.path[1:])

    def to_root_path(self) -> str:
        return _join_s3_paths(self.s3BucketName, self.rootPrefix)

    def to_s3_root_uri(self) -> str:
        return f"s3://{self.to_root_path()}"

    def full_cas_prefix(self) -> str:
        self._validate_root_prefix()
        return _join_s3_paths(self.rootPrefix, S3_DATA_FOLDER_NAME)

    def full_job_output_prefix(self, farm_id, queue_id, job_id) -> str:
        self._validate_root_prefix()
        return _join_s3_paths(self.rootPrefix, S3_MANIFEST_FOLDER_NAME, farm_id, queue_id, job_id)

    def full_step_output_prefix(self, farm_id, queue_id, job_id, step_id) -> str:
        self._validate_root_prefix()
        return _join_s3_paths(
            self.rootPrefix, S3_MANIFEST_FOLDER_NAME, farm_id, queue_id, job_id, step_id
        )

    def full_task_output_prefix(self, farm_id, queue_id, job_id, step_id, task_id) -> str:
        self._validate_root_prefix()
        return _join_s3_paths(
            self.rootPrefix, S3_MANIFEST_FOLDER_NAME, farm_id, queue_id, job_id, step_id, task_id
        )

    def full_output_prefix(
        self, farm_id, queue_id, job_id, step_id, task_id, session_action_id
    ) -> str:
        self._validate_root_prefix()
        return _join_s3_paths(
            self.rootPrefix,
            S3_MANIFEST_FOLDER_NAME,
            farm_id,
            queue_id,
            job_id,
            step_id,
            task_id,
            session_action_id,
        )

    @staticmethod
    def partial_session_action_manifest_prefix(
        farm_id: str,
        queue_id: str,
        job_id: str,
        step_id: str,
        task_id: str,
        session_action_id: str,
        time: float,
    ) -> str:
        """
        Constructs the partial S3 prefix for storing session action output manifests.

        This method creates a hierarchical path structure for organizing output manifests in S3,
        following the pattern: farm_id/queue_id/job_id/step_id/task_id/timestamp_session_action_id.
        The timestamp is converted from a float to an ISO datetime string format.
        """
        return _join_s3_paths(
            farm_id,
            queue_id,
            job_id,
            step_id,
            task_id,
            f"{_float_to_iso_datetime_string(time)}_{session_action_id}",
        )

    @staticmethod
    def partial_session_action_manifest_prefix_without_task(
        farm_id: str,
        queue_id: str,
        job_id: str,
        step_id: str,
        session_action_id: str,
        time: float,
    ) -> str:
        """
        Constructs the partial S3 prefix for storing session action output manifests.

        This method creates a hierarchical path structure for organizing output manifests in S3,
        following the pattern: farm_id/queue_id/job_id/step_id/timestamp_session_action_id.
        The timestamp is converted from a float to an ISO datetime string format.
        """
        return _join_s3_paths(
            farm_id,
            queue_id,
            job_id,
            step_id,
            f"{_float_to_iso_datetime_string(time)}_{session_action_id}",
        )

    def partial_manifest_prefix(self, farm_id, queue_id) -> str:
        guid = _generate_random_guid()
        return _join_s3_paths(
            farm_id,
            queue_id,
            S3_INPUT_MANIFEST_FOLDER_NAME,
            guid,
        )

    def add_root_and_manifest_folder_prefix(self, path: str) -> str:
        """
        Adds “{self.rootPrefix}/{S3_MANIFEST_FOLDER_NAME}/” to the beginning
        of the path and returns it.
        """
        self._validate_root_prefix()
        return _join_s3_paths(self.rootPrefix, S3_MANIFEST_FOLDER_NAME, path)

    def _validate_root_prefix(self) -> None:
        if not self.rootPrefix:
            raise MissingS3RootPrefixError("Missing S3 root prefix")


@dataclass
class Fleet:
    """DataClass to store fleet objects"""

    fleetId: str  # pylint: disable=invalid-name
    priority: int


@dataclass
class Queue:
    """DataClass to store queue objects"""

    queueId: str  # pylint: disable=invalid-name
    displayName: str
    farmId: str  # pylint: disable=invalid-name
    status: str
    defaultBudgetAction: str
    jobAttachmentSettings: Optional[JobAttachmentS3Settings] = None  # pylint: disable=invalid-name


@dataclass
class Job:
    """A non-exhaustive DataClass to store job objects"""

    jobId: str
    attachments: Optional[Attachments] = None  # pylint: disable=invalid-name


@dataclass
class StorageProfile:
    """DataClass to store Storage Profile For Queue objects"""

    storageProfileId: str
    displayName: str
    osFamily: StorageProfileOperatingSystemFamily
    fileSystemLocations: List[FileSystemLocation] = field(default_factory=list)  # type: ignore

    def to_dict(self) -> dict[str, Any]:
        return {
            "storageProfileId": self.storageProfileId,
            "displayName": self.displayName,
            "osFamily": self.osFamily.value,
            "fileSystemLocations": [item.to_dict() for item in self.fileSystemLocations],
        }


@dataclass
class FileSystemLocation:
    """DataClass to store File System Location objects"""

    name: str
    path: str
    type: FileSystemLocationType

    def to_dict(self) -> dict[str, Any]:
        return {"name": self.name, "path": self.path, "type": self.type.value}


class FileSystemLocationType(str, Enum):
    SHARED = "SHARED"
    LOCAL = "LOCAL"


class FileConflictResolution(Enum):
    NOT_SELECTED = 0
    SKIP = 1
    OVERWRITE = 2
    CREATE_COPY = 3


def default_glob_all() -> List[str]:
    return ["**/*"]


@dataclass
class GlobConfig:
    """Include and Exclude configuration for glob input files"""

    include_glob: List[str] = field(default_factory=default_glob_all)
    exclude_glob: List[str] = field(default_factory=list)

    INCLUDE = "include"
    EXCLUDE = "exclude"


@dataclass
class ManifestSnapshot:
    """Data structure to store the results of a manifest snapshot"""

    root: str
    manifest: str


@dataclass
class ManifestDiff:
    """Data structure to store new, modified, or deleted files when comparing manifest to a local file system"""

    new: List[str] = field(default_factory=list)
    modified: List[str] = field(default_factory=list)
    deleted: List[str] = field(default_factory=list)


@dataclass
class ManifestDownload:
    """Data structure to store the S3 and local paths of a manifest"""

    manifest_root: str = field(default_factory=str)
    local_manifest_path: str = field(default_factory=str)


@dataclass
class ManifestMerge:
    """Data structure to store the S3 and local paths of a manifest"""

    manifest_root: str = field(default_factory=str)
    local_manifest_path: str = field(default_factory=str)


@dataclass
class ManifestDownloadResponse:
    """Data structure to capture the response for manifest download"""

    downloaded: list[ManifestDownload] = field(default_factory=list)


@dataclass
class PathMappingRule:
    source_path_format: str
    """The path format associated with the source path (windows vs posix)"""

    source_path: str
    """The path we're looking to change"""

    destination_path: str
    """The path to transform the source path to"""

    def get_hashed_source_path(self, hash_alg: HashAlgorithm) -> str:
        return hash_data(self.source_path.encode("utf-8"), hash_alg)


class FileStatus(Enum):
    """
    Status of local files compared to manifest listed files, comparing hash and time modfied
    """

    UNCHANGED = 0
    NEW = 1
    MODIFIED = 2
    DELETED = 3


@dataclass
class UploadManifestInfo:
    """
    Structured class for output manifest information.

    Attributes:
        output_manifest_path: The relative path to the uploaded output manifest without root prefix
        output_manifest_hash: The hash of the output manifest content
        source_path: Optional source path from the mapping rule
    """

    output_manifest_path: str
    output_manifest_hash: str
    source_path: Optional[str] = None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
The `deadline handle-web-url` command.
"""

import sys
import urllib

import click

from ...config import config_file
from ...exceptions import DeadlineOperationError
from .._common import (
    _PROMPT_WHEN_COMPLETE,
    _handle_error,
    _prompt_at_completion,
)
from .._deadline_web_url import (
    DEADLINE_URL_SCHEME_NAME,
    install_deadline_web_url_handler,
    parse_query_string,
    uninstall_deadline_web_url_handler,
    validate_resource_ids,
)
from .._main import deadline as main
from .job_group import _download_job_output


@main.command(name="handle-web-url")
@click.argument("url", required=False)
@click.option(
    "--prompt-when-complete",
    default=False,
    is_flag=True,
    help="Prompt for keyboard input when completed.",
)
@click.option(
    "--install",
    default=False,
    is_flag=True,
    help=f"Install this CLI command as the {DEADLINE_URL_SCHEME_NAME}:// URL handler",
)
@click.option(
    "--uninstall",
    default=False,
    is_flag=True,
    help=f"Uninstall this CLI command as the {DEADLINE_URL_SCHEME_NAME}:// URL handler",
)
@click.option(
    "--all-users",
    default=False,
    is_flag=True,
    help="With --install or --uninstall, the installation is for all users.",
)
@_handle_error
@click.pass_context
def cli_handle_web_url(
    ctx: click.Context,
    url: str,
    prompt_when_complete: bool,
    install: bool,
    uninstall: bool,
    all_users: bool,
):
    """
    Runs commands sent from a web application such as [Deadline Cloud monitor].

    Use the `--install` and `--uninstall` options to set or remove the command as the `deadline://`
    protocol handler in your operating system.

    Commands use the format `deadline://<handle-web-url-command>?args=value&args=value`.

    This function automatically picks the best AWS profile to use
    based on the provided farm-id and queue-id.

    \b
        deadline://download-output
            ?farm-id=<farm-id>
            &queue-id=<queue-id>
            &job-id=<job-id>
            &step-id=<step-id>                      # optional
            &task-id=<task-id>                      # optional

    [Deadline Cloud monitor]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/working-with-deadline-monitor.html
    """
    ctx.obj[_PROMPT_WHEN_COMPLETE] = prompt_when_complete

    # Determine whether we're handling a URL or installing/uninstalling the AWS Deadline Cloud URL handler
    if url:
        if install or uninstall or all_users:
            raise DeadlineOperationError(
                "The --install, --uninstall and --all-users options cannot be used with a provided URL."
            )

        split_url = urllib.parse.urlsplit(url)

        if split_url.scheme != DEADLINE_URL_SCHEME_NAME:
            raise DeadlineOperationError(
                f"URL scheme {split_url.scheme} is not supported. Only {DEADLINE_URL_SCHEME_NAME} is supported."
            )

        # Validate that the command is supp
        if split_url.netloc == "download-output":
            url_queries = parse_query_string(
                split_url.query,
                parameter_names=["farm-id", "queue-id", "job-id", "step-id", "task-id", "profile"],
                required_parameter_names=["farm-id", "queue-id", "job-id"],
            )

            # Validate the IDs
            # We copy the dict without the 'profile' key as that isn't a resource ID
            validate_resource_ids({k: url_queries[k] for k in url_queries.keys() - {"profile"}})

            farm_id = url_queries.pop("farm_id")
            queue_id = url_queries.pop("queue_id")
            job_id = url_queries.pop("job_id")
            step_id = url_queries.pop("step_id", None)
            task_id = url_queries.pop("task_id", None)

            # Add the standard option "profile", using the one provided by the url (set by Deadline Cloud monitor)
            # or choosing a best guess based on farm and queue IDs
            aws_profile_name = url_queries.pop(
                "profile",
                config_file.get_best_profile_for_farm(farm_id, queue_id),
            )

            # Read the config, and switch the in-memory version to use the chosen AWS profile
            config = config_file.read_config()
            config_file.set_setting("defaults.aws_profile_name", aws_profile_name, config=config)

            _download_job_output(config, farm_id, queue_id, job_id, step_id, task_id)
        else:
            raise DeadlineOperationError(
                f"Command {split_url.netloc} is not supported through handle-web-url.",
            )
    elif install and uninstall:
        raise DeadlineOperationError(
            "Only one of the --install and --uninstall options may be provided."
        )
    elif install:
        install_deadline_web_url_handler(all_users=all_users)
    elif uninstall:
        uninstall_deadline_web_url_handler(all_users=all_users)
    else:
        raise DeadlineOperationError(
            "At least one of a URL, --install, or --uninstall must be provided."
        )

    _prompt_at_completion(ctx)
    sys.exit(0)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import datetime
from functools import wraps
from hashlib import shake_256
from pathlib import Path
import random
import time
from typing import Any, Callable, Optional, Tuple, Type, Union
import uuid
import sys

__all__ = [
    "_join_s3_paths",
    "_generate_random_guid",
    "_float_to_iso_datetime_string",
    "_get_unique_dest_dir_name",
    "_get_bucket_and_object_key",
    "_is_relative_to",
]


TEMP_DOWNLOAD_ADDED_CHARS_LENGTH = 9
"""
Add 9 to path length to account for .Hex value when file is in the middle of downloading in windows.
e.g. test.txt when downloaded becomes test.txt.H4SD9Ddj
"""

WINDOWS_MAX_PATH_LENGTH = 260
"""
Windows Max path length limit of 260.
https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
"""

WINDOWS_UNC_PATH_STRING_PREFIX = "\\\\?\\"
"""
When this is prepended to any path on Windows,
it becomes a UNC path and is allowed to go over the 260 max path length limit.
"""


def _join_s3_paths(root: str, *args: str):
    return "/".join([root, *args])


def _generate_random_guid():
    return str(uuid.uuid4()).replace("-", "")


def _float_to_iso_datetime_string(time: float):
    seconds = int(time)
    microseconds = int((time - seconds) * 1000000)

    dt = datetime.datetime.utcfromtimestamp(seconds) + datetime.timedelta(microseconds=microseconds)
    iso_string = dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    return iso_string


def _get_unique_dest_dir_name(source_root: str) -> str:
    # Note: this is a quick naive way to attempt to prevent colliding
    # relative paths across manifests without adding too much
    # length to the filepaths. length = 2n where n is the number
    # passed to hexdigest.
    return f"assetroot-{shake_256(source_root.encode()).hexdigest(10)}"


def _get_bucket_and_object_key(s3_path: str) -> Tuple[str, str]:
    """Returns the bucket name and object key from the S3 URI"""
    bucket, key = s3_path.replace("s3://", "").split("/", maxsplit=1)
    return bucket, key


def _normalize_windows_path(path: Union[Path, str]) -> Path:
    """
    Strips \\\\?\\ prefix from Windows paths.
    """
    p_str = str(path)
    if p_str.startswith("\\\\?\\"):
        return Path(p_str[4:])
    return Path(path)


def _is_relative_to(path1: Union[Path, str], path2: Union[Path, str]) -> bool:
    """
    Determines if path1 is relative to path2. This function is to support
    Python versions (3.7 and 3.8) that do not have the built-in `Path.is_relative_to()` method.
    """
    try:
        p1 = _normalize_windows_path(Path(path1).resolve())
        p2 = _normalize_windows_path(Path(path2).resolve())
        p1.relative_to(p2)
        return True
    except ValueError:
        return False


def _is_windows_long_path_registry_enabled() -> bool:
    if sys.platform != "win32":
        return True

    import ctypes

    ntdll = ctypes.WinDLL("ntdll")
    ntdll.RtlAreLongPathsEnabled.restype = ctypes.c_ubyte
    ntdll.RtlAreLongPathsEnabled.argtypes = ()

    return bool(ntdll.RtlAreLongPathsEnabled())


def _get_long_path_compatible_path(original_path: Union[str, Path]) -> Path:
    """
    Given a Path or string representing a path,
    make it long path compatible if needed on Windows and return the Path object
    https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation

    :param original_path: Original unmodified path/string representing an absolute path.
    show
    :param show_long_path_warning: Whether to show a warning to the user that the resulting path is in a long path.
    :return: A Path object representing the long path compatible path.
    """

    original_path_string = str(original_path)
    if sys.platform != "win32":
        return Path(original_path_string)

    if (
        len(original_path_string) + TEMP_DOWNLOAD_ADDED_CHARS_LENGTH >= WINDOWS_MAX_PATH_LENGTH
        and not original_path_string.startswith(WINDOWS_UNC_PATH_STRING_PREFIX)
        and not _is_windows_long_path_registry_enabled()
    ):
        # Prepend \\?\ to the file name to treat it as an UNC path
        return Path(WINDOWS_UNC_PATH_STRING_PREFIX + original_path_string)
    return Path(original_path_string)


def _retry(
    ExceptionToCheck: Union[Type[Exception], Tuple[Type[Exception], ...]] = AssertionError,
    tries: int = 2,
    delay: Union[int, float, Tuple[Union[int, float], Union[int, float]]] = 1.0,
    backoff: float = 1.0,
    logger: Optional[Callable] = print,
) -> Callable:
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: float or tuple
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: float
    :param logger: logging function to use. If None, won't log
    :type logger: logging.Logger instance
    """

    def deco_retry(f: Callable) -> Callable:
        @wraps(f)
        def f_retry(*args: Any, **kwargs: Any) -> Callable:
            mtries: int = tries
            if isinstance(delay, (float, int)):
                mdelay = delay
            elif isinstance(delay, tuple):
                mdelay = random.uniform(delay[0], delay[1])
            else:
                raise ValueError(f"Provided delay {delay} isn't supported")

            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    if logger:
                        logger(f"{str(e)}, Retrying in {mdelay} seconds...")
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Module that defines the asset manifest versions."""

from enum import Enum


class ManifestVersion(str, Enum):
    """
    Enumerant of all Asset Manifest versions supported by this library.

    Special values:
      UNDEFINED -- Purely for internal testing.

    Versions:
      v2023_03_03 - First version.
    """

    UNDEFINED = "UNDEFINED"
    v2023_03_03 = "2023-03-03"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Exceptions for the AWS Deadline Cloud Client Library.
"""


class DeadlineOperationError(Exception):
    """Error whose message gets printed verbatim by the cli handler"""


class DeadlineOperationCanceled(DeadlineOperationError):
    """DeadlineOperationError for when an operation was canceled"""

    def __init__(self, message: str = "Operation canceled"):
        super().__init__(message)


class DeadlineOperationTimedOut(DeadlineOperationError):
    """DeadlineOperationError for when an operation timed out"""

    def __init__(self, message: str = "Operation timed out"):
        super().__init__(message)


class CreateJobWaiterCanceled(DeadlineOperationCanceled):
    """Error for when the waiter after CreateJob is interrupted"""

    def __init__(self, message: str = "Operation canceled while waiting for CreateJob to finish"):
        super().__init__(message)


class UserInitiatedCancel(DeadlineOperationCanceled):
    """Error for when the user requests cancelation"""

    def __init__(self, message: str = "Operation canceled by user"):
        super().__init__(message)


class NonValidInputError(Exception):
    """Error for when the user input is nonvalid"""


class ManifestOutdatedError(Exception):
    """Error for when local files are different from version captured in manifest"""



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

import json
import os
from importlib import reload
from logging import getLogger
from typing import Any, Dict, Optional
import copy

from qtpy.QtCore import Qt  # pylint: disable=import-error
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QMainWindow,
    QWidget,
)

from ..job_bundle import deadline_yaml_dump
from ..job_bundle.parameters import JobParameter
from .dataclasses import CliJobSettings
from .dialogs.submit_job_to_deadline_dialog import (
    SubmitJobToDeadlineDialog,
    JobBundlePurpose,
)
from .widgets.cli_job_settings_tab import CliJobSettingsWidget
from ..job_bundle.submission import AssetReferences

logger = getLogger(__name__)

# This contains the dialog we create, so that we only open one submitter dialog
__submitter_dialog = None


def show_cli_job_submitter(parent: Optional[QWidget] = None, f=Qt.WindowFlags()) -> None:
    """
    Shows an example CLI Job Submitter.

    Pass f=Qt.Tool if running it within an application context and want it
    to stay on top.
    """
    global __submitter_dialog

    close_submitter()

    if parent is None:
        # Get the main application window so we can parent ours to it
        app = QApplication.instance()
        parent = [widget for widget in app.topLevelWidgets() if isinstance(widget, QMainWindow)][0]  # type: ignore[union-attr]

    def on_create_job_bundle_callback(
        widget: SubmitJobToDeadlineDialog,
        job_bundle_dir: str,
        settings: CliJobSettings,
        queue_parameters: list[JobParameter],
        asset_references: AssetReferences,
        host_requirements: Optional[Dict[str, Any]] = None,
        purpose: JobBundlePurpose = JobBundlePurpose.SUBMISSION,
    ) -> dict[str, Any]:
        """
        Perform a submission when the submit button is pressed

        Args:
            widget (SubmitJobToDeadlineDialog): The Deadline job submission dialog.
            settings (CliJobSettings): A settings object that was populated from the job submission dialog.
            job_bundle_dir (str): The directory within which to create the job bundle.
            asset_references (FlatAssetReferences): The input from the attachments provided during
                construction and the user's input in the Job Attachments tab.
        """
        if settings.file_format not in ("YAML", "JSON"):
            raise RuntimeError(
                f"The CLI Job Submitter only supports YAML and JSON output, not {settings.file_format!r}."
            )

        job_template: Dict[str, Any] = {
            "specificationVersion": "jobtemplate-2023-09",
            "name": settings.name,
        }
        if settings.description:
            job_template["description"] = settings.description
        job_template["parameterDefinitions"] = [
            {
                "name": "DataDir",
                "type": "PATH",
                "objectType": "DIRECTORY",
                "dataFlow": "INOUT",
            }
        ]
        step = {
            "name": "CliScript",
            "script": {
                "actions": {
                    "onRun": {
                        "command": "{{Task.File.runScript}}",
                    }
                },
                "embeddedFiles": [
                    {
                        "name": "runScript",
                        "type": "TEXT",
                        "runnable": True,
                        "data": settings.bash_script_contents,
                    }
                ],
            },
        }
        job_template["steps"] = [step]
        if settings.use_array_parameter:
            job_array_parameter_name = f"{settings.array_parameter_name}Values"
            job_template["parameterDefinitions"].append(
                {
                    "name": job_array_parameter_name,
                    "type": "STRING",
                    "default": settings.array_parameter_values,
                }
            )
            step["parameterSpace"] = {
                "taskParameterDefinitions": [
                    {
                        "name": settings.array_parameter_name,
                        "type": "INT",
                        "range": "{{Param." + job_array_parameter_name + "}}",
                    }
                ]
            }

        # If "HostRequirements" is provided, inject it into each of the "Step"
        if host_requirements:
            # for each step in the template, append the same host requirements.
            for step in job_template["steps"]:
                step["hostRequirements"] = copy.deepcopy(host_requirements)

        with open(
            os.path.join(job_bundle_dir, f"template.{settings.file_format.lower()}"),
            "w",
            encoding="utf8",
        ) as f:
            if settings.file_format == "YAML":
                deadline_yaml_dump(job_template, f)
            elif settings.file_format == "JSON":
                json.dump(job_template, f, sort_keys=False, indent=1)

        # Filter the provided queue parameters to just their values
        parameters_values = [
            {"name": param["name"], "value": param["value"]} for param in queue_parameters
        ]

        with open(
            os.path.join(job_bundle_dir, f"parameter_values.{settings.file_format.lower()}"),
            "w",
            encoding="utf8",
        ) as f:
            if settings.file_format == "YAML":
                deadline_yaml_dump({"parameterValues": parameters_values}, f)
            elif settings.file_format == "JSON":
                json.dump({"parameterValues": parameters_values}, f, indent=1)

        if asset_references:
            with open(
                os.path.join(job_bundle_dir, f"asset_references.{settings.file_format.lower()}"),
                "w",
                encoding="utf8",
            ) as f:
                if settings.file_format == "YAML":
                    deadline_yaml_dump(asset_references.to_dict(), f)
                elif settings.file_format == "JSON":
                    json.dump(asset_references.to_dict(), f, indent=1)

        return {
            "known_asset_paths": [settings.data_dir],
            "job_parameters": [{"name": "DataDir", "value": settings.data_dir}],
        }

    __submitter_dialog = SubmitJobToDeadlineDialog(
        job_setup_widget_type=CliJobSettingsWidget,
        initial_job_settings=CliJobSettings(),
        initial_shared_parameter_values={},
        auto_detected_attachments=AssetReferences(),
        attachments=AssetReferences(),
        show_host_requirements_tab=True,
        on_create_job_bundle_callback=on_create_job_bundle_callback,
        parent=parent,
        f=f,
    )
    __submitter_dialog.show()


def close_submitter() -> None:
    global __submitter_dialog

    if __submitter_dialog:
        __submitter_dialog.close()
        __submitter_dialog = None


def _reload_modules(mod):
    # TODO: Put this code where it makes sense
    """
    Recursively reloads all modules in the specified package, in postfix order
    """
    import types

    child_mods = [
        m
        for m in mod.__dict__.values()
        if isinstance(m, types.ModuleType)
        and m.__package__
        and m.__package__.startswith(mod.__package__)
    ]

    for child in child_mods:
        _reload_modules(mod=child)

    reload(mod)


def reload_plugin() -> None:
    # TODO: Put this code where it makes sense
    """
    Designed for interative development without closing the DCC
    the submitter is embedded inside. Closes the submitter,
    reloads the Python modules, and starts the submitter again.
    """
    close_submitter()

    # Reload the AWS Deadline Cloud submitter code
    import deadline

    _reload_modules(deadline)

    # Re-open the submitter
    show_cli_job_submitter()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline auth` commands:
    * login
    * logout
    * status
"""

import click
import json
import logging

from ... import api
from .._main import deadline as main
from ...api._session import _modified_logging_level, AwsCredentialsSource
from ...config import config_file, get_setting
from .._common import _apply_cli_options_to_config, _handle_error

JSON_FIELD_PROFILE_NAME = "profile_name"
JSON_FIELD_AUTH_STATUS = "status"
JSON_FIELD_CREDS_SOURCE = "source"
JSON_FIELD_AUTH_API_AVAILABLE = "api_availability"


def _cli_on_pending_authorization(**kwargs):
    """
    Callback for `login`, to tell the user that Deadline Cloud monitor is opening
    """

    if kwargs["credentials_source"] == AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN:
        click.echo("Opening Deadline Cloud monitor. Please log in and then return here.")


@main.group(name="auth")
@_handle_error
def cli_auth():
    """
    Commands to handle authentication, such as logging in to [Deadline Cloud monitor].

    [Deadline Cloud monitor]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/working-with-deadline-monitor.html
    """


@cli_auth.command(name="login")
@_handle_error
def auth_login():
    """
    Opens [Deadline Cloud monitor] to log in to your farm. Supports
    profiles created by Deadline Cloud monitor.

    [Deadline Cloud monitor]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/working-with-deadline-monitor.html

    """
    click.echo(
        f"Logging into AWS Profile {config_file.get_setting('defaults.aws_profile_name')!r} for AWS Deadline Cloud"
    )

    message = api.login(
        on_pending_authorization=_cli_on_pending_authorization, on_cancellation_check=None
    )

    click.echo(f"\nSuccessfully logged in: {message}\n")


@cli_auth.command(name="logout")
@_handle_error
def auth_logout():
    """
    Logs out of the configured AWS profile, if it was created by Deadline Cloud monitor.
    """
    api.logout()

    click.echo("Successfully logged out of all Deadline Cloud monitor AWS profiles")


@cli_auth.command(name="status")
@click.option("--profile", help="The AWS profile to use.")
@click.option(
    "--output",
    type=click.Choice(
        ["verbose", "json"],
        case_sensitive=False,
    ),
    default="verbose",
    help="Specifies the output format of the messages printed to stdout.\n"
    "VERBOSE: Displays messages in a human-readable text format.\n"
    "JSON: Displays messages in JSON line format, so that the info can be easily "
    "parsed/consumed by custom scripts.",
)
@_handle_error
def auth_status(output, **args):
    """Gets the status of the selected AWS profile, including its name, whether it was created by
    [Deadline Cloud monitor], and whether
    [Deadline Cloud APIs] are accessible.

    [Deadline Cloud monitor]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/working-with-deadline-monitor.html
    [Deadline Cloud APIs]: https://docs.aws.amazon.com/deadline-cloud/latest/APIReference/Welcome.html
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(**args)
    profile_name = get_setting("defaults.aws_profile_name", config=config)
    is_json_format = True if output == "json" else False

    with _modified_logging_level(logging.getLogger("deadline.client.api"), logging.CRITICAL):
        # always returns enum in AwsCredentialsSource
        creds_source = api.get_credentials_source(config=config)
        creds_source_result = creds_source.name

        # always returns enum in AwsAuthenticationStatus
        auth_status = api.check_authentication_status(config=config)
        auth_status_results = auth_status.name

        # always returns True/False
        api_availability_result = api.check_deadline_api_available(config=config)

    if not is_json_format:
        width = 17
        click.echo(f"{'Profile Name:': >{width}} {profile_name}")
        click.echo(f"{'Source:': >{width}} {creds_source_result}")
        click.echo(f"{'Status:': >{width}} {auth_status_results}")
        click.echo(f"{'API Availability:': >{width}} {api_availability_result}")
    else:
        json_output = {
            JSON_FIELD_PROFILE_NAME: profile_name,
            JSON_FIELD_CREDS_SOURCE: creds_source_result,
            JSON_FIELD_AUTH_STATUS: auth_status_results,
            JSON_FIELD_AUTH_API_AVAILABLE: api_availability_result,
        }
        click.echo(json.dumps(json_output, ensure_ascii=True))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline bundle` commands.
"""

from __future__ import annotations

import json
import logging
import sys
import re
from typing import Any, Optional
import tempfile
import shutil
import os
from dataclasses import fields

import click
from botocore.exceptions import ClientError

from ... import api
from ...config import config_file
from ...dataclasses import SubmitterInfo
from ....job_attachments.exceptions import (
    AssetSyncError,
    AssetSyncCancelledError,
    MisconfiguredInputsError,
)
from ....job_attachments.models import JobAttachmentsFileSystem

from ...exceptions import DeadlineOperationError, CreateJobWaiterCanceled
from .._common import (
    _apply_cli_options_to_config,
    _handle_error,
    _ProgressBarCallbackManager,
    _parse_multi_format_parameters,
    _suggest_resources_on_client_error,
)
from .._main import deadline as main
from ._sigint_handler import SigIntHandler

logger = logging.getLogger(__name__)

# Set up the signal handler for handling Ctrl + C interruptions.
sigint_handler = SigIntHandler()


@main.group(name="bundle")
@_handle_error
def cli_bundle():
    """
    Commands to work with Open Job Description [job bundles]. Use these commands to
    submit jobs to run on a Deadline Cloud queue.

    [job bundles]: https://docs.aws.amazon.com/deadline-cloud/latest/developerguide/build-job-bundle.html
    """


# Latin alphanumeric, starting with a letter
_openjd_identifier_regex = r"(?-m:^[A-Za-z_][A-Za-z0-9_]*\Z)"


def validate_parameters(ctx, param, value):
    """
    Validate provided --parameter values, ensuring that they are in the format "ParamName=Value", and convert them to a dict with the
    following format:
        [{"name": "<name>", "value": "<value>"}, ...]
    """
    parameters_split = []
    for parameter in value:
        regex_match = re.match("([^=]+)=(.*)", parameter)
        if not regex_match:
            raise click.BadParameter(
                f'Parameters must be provided in the format "ParamName=Value". Invalid parameter: {parameter}'
            )

        if not re.match(_openjd_identifier_regex, regex_match[1]):
            raise click.BadParameter(
                f"Parameter names must be alphanumeric Open Job Description identifiers. Invalid parameter name: {regex_match[1]}"
            )

        parameters_split.append({"name": regex_match[1], "value": regex_match[2]})

    return parameters_split


def _validate_submitter_info(ctx, param, values):
    """
    Validate provided --submitter-info value and convert to SubmitterInfo object.

    Supports three input formats that can be mixed:
    - Key=value pairs: --submitter-info submitter_name=MyApp --submitter-info host_application_name=Maya
    - Inline JSON strings: --submitter-info '{"submitter_name": "MyApp", "additional_info": {"custom": "data"}}'
    - File paths (JSON or YAML): --submitter-info file://path/to/submitter.json

    All keys must be valid SubmitterInfo fields. Unknown keys will raise an error.
    """
    if not values:
        return None

    # Get valid field names from SubmitterInfo dataclass
    valid_fields = {field.name for field in fields(SubmitterInfo)}

    info_dict = _parse_multi_format_parameters(list(values))

    # Validate all keys
    for key in info_dict.keys():
        if key not in valid_fields:
            raise click.BadParameter(
                f"Unknown field '{key}'. Valid fields are: {', '.join(sorted(valid_fields))}"
            )

    # Ensure submitter_name is provided as a required field
    if "submitter_name" not in info_dict:
        raise click.BadParameter(
            "submitter_name is required when using --submitter-info. "
            "Example: --submitter-info submitter_name=MyApp"
        )

    try:
        return SubmitterInfo(**info_dict)
    except TypeError as e:
        raise click.BadParameter(f"Failed to create SubmitterInfo: {e}") from e


def _interactive_confirmation_prompt(message: str, default_response: bool) -> bool:
    """
    Callback to decide if submission should continue or be canceled. Returns True to continue, False to cancel.

    Args:
        warning_message (str): The warning message to display.
        default_response (bool): The default to present as the response (True to continue, False to cancel).
    """
    return click.confirm(
        message,
        default=default_response,
    )


@cli_bundle.command(name="submit")
@click.option(
    "-p",
    "--parameter",
    multiple=True,
    callback=validate_parameters,
    help=(
        "The values for the job template's parameters. Can be provided as key-value pairs, inline JSON strings, "
        "or as paths to a JSON or YAML document. Later values for repeated parameter names take precedence. "
        'Examples: --parameter MyParam=5 -p file://parameter_file.json -p \'{"OtherParam": "10"}\''
    ),
)
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--storage-profile-id", help="The storage profile to use.")
@click.option("--name", help="The job name to use in place of the one in the job bundle.")
@click.option(
    "--priority",
    type=int,
    default=50,
    help="The priority of the job. Jobs with a higher priority run first.",
)
@click.option(
    "--max-failed-tasks-count",
    type=int,
    help="The maximum number of failed tasks before the job is marked as failed.",
)
@click.option(
    "--max-retries-per-task",
    type=int,
    help="The maximum number of times to retry a task before it is marked as failed.",
)
@click.option(
    "--max-worker-count",
    type=int,
    help="The max worker count of the job.",
)
@click.option(
    "--target-task-run-status",
    type=click.Choice(["READY", "SUSPENDED"], case_sensitive=False),
    help="The target task run status for the job. READY means tasks will start immediately, "
    "SUSPENDED means tasks will be created but not start until manually resumed.",
)
@click.option(
    "--job-attachments-file-system",
    help="The method workers use to access job attachments. "
    "COPIED means to copy files to the worker and VIRTUAL means to load "
    "files as needed from a virtual file system. If VIRTUAL is selected "
    "but not supported by a worker, it will fallback to COPIED.",
    type=click.Choice([e.value for e in JobAttachmentsFileSystem]),
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@click.option(
    "--require-paths-exist",
    is_flag=True,
    help="Return an error if any input files are missing.",
)
@click.option(
    "--submitter-name",
    type=click.STRING,
    help="Name of the application submitting the bundle.",
)
@click.option(
    "--known-asset-path",
    multiple=True,
    help="Path that should not generate warnings when outside storage profile locations. "
    "Can be specified multiple times for different paths.",
)
@click.option(
    "--save-debug-snapshot",
    help="EXPERIMENTAL - Instead of submitting the job, generate a debug snapshot as a directory or a zip file if the extension is .zip."
    " It includes the job attachments and parameters for creating the job."
    " You can later run the bash script in the snapshot to submit the job using AWS CLI commands.",
)
@click.option(
    "--force-s3-check/--no-force-s3-check",
    default=None,
    help="Force verification that job attachments exist in S3 before skipping upload. "
    "Use when S3 bucket contents may be out of sync with local caches. "
    "Overrides the 'settings.force_s3_check' config setting.",
)
@click.argument("job_bundle_dir")
@_handle_error
def bundle_submit(
    job_bundle_dir,
    job_attachments_file_system,
    parameter,
    known_asset_path,
    name,
    priority,
    max_failed_tasks_count,
    max_retries_per_task,
    max_worker_count,
    target_task_run_status,
    require_paths_exist,
    submitter_name,
    save_debug_snapshot,
    force_s3_check,
    **args,
):
    """
    Submits an Open Job Description [job bundle] to a
    [Deadline Cloud queue]. You can provide options
    to set parameter values, the job name, priority, and more.

    [job bundle]: https://docs.aws.amazon.com/deadline-cloud/latest/developerguide/build-job-bundle.html
    [Deadline Cloud queue]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/queues.html
    """
    # Apply the CLI args to the config
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    # Resolve force_s3_check: CLI flag takes precedence, otherwise use config setting
    if force_s3_check is None:
        force_s3_check = config_file.str2bool(
            config_file.get_setting("settings.force_s3_check", config=config)
        )

    hash_callback_manager = _ProgressBarCallbackManager(length=100, label="Hashing Attachments")
    upload_callback_manager = _ProgressBarCallbackManager(length=100, label="Uploading Attachments")

    def _check_create_job_wait_canceled() -> bool:
        return sigint_handler.continue_operation

    try:
        snapshot_tmpdir = None
        if save_debug_snapshot:
            save_debug_snapshot = os.path.abspath(save_debug_snapshot)

            # If the debug snapshot is to a zip file, first put it in a temporary directory
            if save_debug_snapshot.endswith(".zip"):
                snapshot_tmpdir = tempfile.TemporaryDirectory()

        job_id = api.create_job_from_job_bundle(
            job_bundle_dir=job_bundle_dir,
            job_parameters=parameter,
            name=name,
            job_attachments_file_system=job_attachments_file_system,
            config=config,
            priority=priority,
            max_failed_tasks_count=max_failed_tasks_count,
            max_retries_per_task=max_retries_per_task,
            max_worker_count=max_worker_count,
            target_task_run_status=target_task_run_status,
            hashing_progress_callback=hash_callback_manager.callback,
            upload_progress_callback=upload_callback_manager.callback,
            create_job_result_callback=_check_create_job_wait_canceled,
            print_function_callback=click.echo,
            interactive_confirmation_callback=_interactive_confirmation_prompt,
            require_paths_exist=require_paths_exist,
            submitter_name=submitter_name or "CLI",
            known_asset_paths=known_asset_path,
            debug_snapshot_dir=snapshot_tmpdir.name if snapshot_tmpdir else save_debug_snapshot,
            force_s3_check=force_s3_check,
        )

        if snapshot_tmpdir:
            # Put the snapshot in a zip file
            os.makedirs(os.path.dirname(save_debug_snapshot), exist_ok=True)
            shutil.make_archive(save_debug_snapshot, "zip", snapshot_tmpdir.name)

        if save_debug_snapshot:
            click.echo("Saved job debug snapshot:")
            click.echo(f"    {save_debug_snapshot}")

        # Check Whether the CLI options are modifying any of the default settings that affect
        # the job id. If not, we'll save the job id submitted as the default job id.
        # If a job snapshot directory was provided, the job_id will be None.
        if (
            args.get("profile") is None
            and args.get("farm_id") is None
            and args.get("queue_id") is None
            and args.get("storage_profile_id") is None
            and job_id
        ):
            config_file.set_setting("defaults.job_id", job_id)

    except AssetSyncCancelledError as exc:
        if sigint_handler.continue_operation:
            raise DeadlineOperationError(f"Job submission unexpectedly canceled:\n{exc}") from exc
        else:
            click.echo("Job submission canceled.")
            sys.exit(1)
    except AssetSyncError as exc:
        raise DeadlineOperationError(f"Failed to upload job attachments:\n{exc}") from exc
    except CreateJobWaiterCanceled as exc:
        if sigint_handler.continue_operation:
            raise DeadlineOperationError(
                f"Unexpectedly canceled during wait for final status of CreateJob:\n{exc}"
            ) from exc
        else:
            click.echo("Canceled waiting for final status of CreateJob.")
            sys.exit(1)
    except ClientError as exc:
        suggestion = _suggest_resources_on_client_error(
            exc,
            farm_id=config_file.get_setting("defaults.farm_id", config=config),
            queue_id=config_file.get_setting("defaults.queue_id", config=config),
            config=config,
        )
        raise DeadlineOperationError(
            f"Failed to submit the job bundle to AWS Deadline Cloud:\n{exc}{suggestion}"
        ) from exc
    except MisconfiguredInputsError as exc:
        click.echo(str(exc))
        click.echo("Job submission canceled.")
        sys.exit(1)
    except Exception as exc:
        api.get_deadline_cloud_library_telemetry_client().record_error(
            event_details={"exception_scope": "on_submit"},
            exception_type=str(type(exc)),
        )
        raise
    finally:
        if snapshot_tmpdir:
            snapshot_tmpdir.cleanup()


@cli_bundle.command(name="gui-submit")
@click.option(
    "-p",
    "--parameter",
    multiple=True,
    callback=validate_parameters,
    help=(
        "Initial values in the GUI for the job template's parameters. Can be provided as key-value pairs, inline JSON strings, "
        "or as paths to a JSON or YAML document. Later values for repeated parameter names take precedence. "
        'Examples: --parameter MyParam=5 -p file://parameter_file.json -p \'{"OtherParam": "10"}\''
    ),
)
@click.argument("job_bundle_dir", required=False)
@click.option(
    "--browse",
    is_flag=True,
    help="Opens a folder browser to select a bundle.",
)
@click.option(
    "--install-gui",
    is_flag=True,
    help="Installs GUI dependencies if they are not installed already",
)
@click.option(
    "--submitter-name",
    help="[DEPRECATED] Use --submitter-info submitter_name=<name> instead. Name of the application submitting the bundle. If a name is specified, the GUI will automatically close after submitting the job.",
)
@click.option(
    "--output",
    type=click.Choice(
        ["verbose", "json"],
        case_sensitive=False,
    ),
    default="verbose",
    help="Specifies the output format of the messages printed to stdout.\n"
    "VERBOSE: Displays messages in a human-readable text format.\n"
    "JSON: Displays messages in JSON line format, so that the info can be easily "
    "parsed/consumed by custom scripts.",
)
@click.option(
    "--known-asset-path",
    multiple=True,
    help="Path that should not generate warnings when outside storage profile locations. "
    "Can be specified multiple times for different paths.",
)
@click.option(
    "--submitter-info",
    multiple=True,
    callback=_validate_submitter_info,
    help="Submitter and environment information. Supports key=value pairs, inline JSON strings, "
    "and file paths (JSON or YAML). Later values for repeated fields take precedence. "
    "Examples: --submitter-info submitter_name=MyApp --submitter-info host_application_name=Maya "
    'OR --submitter-info \'{"submitter_name": "MyApp", "additional_info": {"render_engine": "Cycles"}}\' '
    "OR --submitter-info file://path/to/submitter.json",
)
@_handle_error
def bundle_gui_submit(
    parameter,
    job_bundle_dir,
    browse,
    output,
    install_gui,
    known_asset_path,
    submitter_name,
    submitter_info,
    **args,
):
    """
    Opens a GUI to submit an Open Job Description [job bundle] to a
    [Deadline Cloud queue]. You can provide options
    to set the initial parameter values shown in the GUI.

    [job bundle]: https://docs.aws.amazon.com/deadline-cloud/latest/developerguide/build-job-bundle.html
    [Deadline Cloud queue]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/queues.html
    """

    if submitter_name:
        click.echo(
            click.style(
                "DeprecationWarning: The option --submitter-name is deprecated. Use --submitter-info instead.",
                fg="red",
            ),
            err=True,
        )
        if submitter_info:
            # --submitter-name takes precedence if we already have submitter_info provided
            submitter_info.submitter_name = submitter_name
        else:
            submitter_info = SubmitterInfo(submitter_name=submitter_name)

    from ...ui import gui_context_for_cli
    from ...ui._utils import tr

    with gui_context_for_cli(automatically_install_dependencies=install_gui) as app:
        from ...ui.job_bundle_submitter import show_job_bundle_submitter

        if not job_bundle_dir and not browse:
            raise DeadlineOperationError(
                tr(
                    "Specify a job bundle directory or run the bundle command with the --browse flag"
                )
            )
        output = output.lower()

        submitter = show_job_bundle_submitter(
            input_job_bundle_dir=job_bundle_dir,
            browse=browse,
            submitter_info=submitter_info,
            known_asset_paths=known_asset_path,
            job_parameters=parameter,
        )

        if not submitter:
            return

        submitter.show()

        app.exec()

        _print_response(
            output=output,
            job_bundle_dir=job_bundle_dir,
            job_history_bundle_dir=submitter.job_history_bundle_dir,
            job_id=submitter.job_id,
        )


def _print_response(
    output: str,
    job_bundle_dir: str,
    job_history_bundle_dir: Optional[str],
    job_id: Optional[str],
):
    if output == "json":
        if job_id:
            response: dict[str, Any] = {
                "status": "SUBMITTED",
                "jobId": job_id,
                "jobHistoryBundleDirectory": job_history_bundle_dir,
            }
            click.echo(json.dumps(response))
        else:
            click.echo(json.dumps({"status": "CANCELED"}))
    else:
        if job_id:
            click.echo("Submitted job bundle:")
            click.echo(f"   {job_bundle_dir}")
            click.echo(f"Job ID: {job_id}")
        else:
            click.echo("Job submission canceled.")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import logging
import os
import shutil
import subprocess
import time
from pathlib import Path
import threading
from typing import Dict, Union, Optional

from .exceptions import (
    VFSExecutableMissingError,
    VFSFailedToMountError,
    VFSLaunchScriptMissingError,
    VFSRunPathNotSetError,
)

from .os_file_permission import PosixFileSystemPermissionSettings

from deadline.client.api import (
    get_deadline_cloud_library_telemetry_client as _get_deadline_cloud_library_telemetry_client,
)

log = logging.getLogger(__name__)

DEADLINE_VFS_ENV_VAR = "DEADLINE_VFS_PATH"
DEADLINE_VFS_CACHE_ENV_VAR = "DEADLINE_VFS_CACHE"
DEADLINE_VFS_EXECUTABLE = "deadline_vfs"
DEADLINE_VFS_INSTALL_PATH = "/opt/deadline_vfs"
DEADLINE_VFS_EXECUTABLE_SCRIPT = "/scripts/production/al2/run_deadline_vfs_al2.sh"

DEADLINE_VFS_PID_FILE_NAME = "vfs_pids.txt"
DEADLINE_MANIFEST_GROUP_READ_PERMS = 0o640

VFS_CACHE_REL_PATH_IN_SESSION = ".vfs_object_cache"
VFS_MANIFEST_FOLDER_IN_SESSION = ".vfs_manifests"
VFS_LOGS_FOLDER_IN_SESSION = ".vfs_logs"

VFS_MANIFEST_FOLDER_PERMISSIONS = PosixFileSystemPermissionSettings(
    os_user="",
    os_group="",
    dir_mode=0o31,
    file_mode=0o64,
)


class VFSProcessManager(object):
    exe_path: Optional[str] = None
    launch_script_path: Optional[str] = None
    library_path: Optional[str] = None
    cwd_path: Optional[str] = None

    _mount_point: str
    _vfs_proc: Optional[subprocess.Popen]
    _vfs_thread: Optional[threading.Thread]
    _mount_temp_directory: Optional[str]
    _run_path: Optional[Union[os.PathLike, str]]
    _asset_bucket: str
    _region: str
    _manifest_path: str
    _os_user: str
    _os_env_vars: Dict[str, str]
    _os_group: Optional[str]
    _cas_prefix: Optional[str]
    _asset_cache_path: Optional[str]

    def __init__(
        self,
        asset_bucket: str,
        region: str,
        manifest_path: str,
        mount_point: str,
        os_user: str,
        os_env_vars: Dict[str, str],
        os_group: Optional[str] = None,
        cas_prefix: Optional[str] = None,
        asset_cache_path: Optional[str] = None,
    ):
        self._mount_point = mount_point
        self._vfs_proc = None
        self._vfs_thread = None
        self._mount_temp_directory = None
        self._run_path = None
        self._asset_bucket = asset_bucket
        self._region = region
        self._manifest_path = manifest_path
        self._os_user = os_user
        self._os_group = os_group
        self._os_env_vars = os_env_vars
        self._cas_prefix = cas_prefix
        self._asset_cache_path = asset_cache_path

    @classmethod
    def kill_all_processes(cls, session_dir: Path, os_user: str) -> None:
        """
        Kill all existing VFS processes when outputs have been uploaded.
        :param session_dir: tmp directory for session
        :param os_user: the user running the job.
        """
        log.info("Terminating all VFS processes.")
        try:
            pid_file_path = (session_dir / DEADLINE_VFS_PID_FILE_NAME).resolve()
            with open(pid_file_path, "r") as file:
                for line in file.readlines():
                    line = line.strip()
                    mount_point, _, _ = line.split(":")
                    cls.shutdown_libfuse_mount(mount_point, os_user, session_dir)
            os.remove(pid_file_path)
        except FileNotFoundError:
            log.warning(f"VFS pid file not found at {pid_file_path}")

    @classmethod
    def get_shutdown_args(cls, mount_path: str, os_user: str):
        """
        Return the argument list to provide the subprocess run command to shut down the mount
        :param mount_path: path to mounted folder
        :param os_user: the user running the job.
        """
        fusermount3_path = os.path.join(cls.find_vfs_link_dir(), "fusermount3")
        if not os.path.exists(fusermount3_path):
            log.warning(f"fusermount3 not found at {cls.find_vfs_link_dir()}")
            return None
        return ["sudo", "-u", os_user, fusermount3_path, "-u", mount_path]

    @classmethod
    def shutdown_libfuse_mount(cls, mount_path: str, os_user: str, session_dir: Path) -> bool:
        """
        Shut down the mount at the provided path using the fusermount3 unmount option
        as the provided user
        :param mount_path: path to mounted folder
        """
        log.info(f"Attempting to shut down {mount_path} as {os_user}")
        shutdown_args = cls.get_shutdown_args(mount_path, os_user)
        if not shutdown_args:
            return False
        try:
            run_result = subprocess.run(shutdown_args, check=True)
        except subprocess.CalledProcessError as e:
            log.warning(f"Shutdown failed with error {e}")
            # Don't reraise, check if mount is gone
        log.info(f"Shutdown returns {run_result.returncode}")
        return cls.wait_for_mount(mount_path, session_dir, expected=False)

    @classmethod
    def kill_process_at_mount(cls, session_dir: Path, mount_point: str, os_user: str) -> bool:
        """
        Kill the VFS instance running at the given mount_point and modify the VFS pid tracking
        file to remove the entry.

        :param session_dir: tmp directory for session
        :param mount_point: local directory to search for
        :param os_user: user to attempt shut down as
        """
        if not cls.is_mount(mount_point):
            log.info(f"{mount_point} is not a mount, returning")
            return False
        log.info(f"Terminating deadline_vfs processes at {mount_point}.")
        mount_point_found: bool = False
        try:
            pid_file_path = (session_dir / DEADLINE_VFS_PID_FILE_NAME).resolve()
            with open(pid_file_path, "r") as file:
                lines = file.readlines()
            with open(pid_file_path, "w") as file:
                for line in lines:
                    line = line.strip()
                    if mount_point_found:
                        file.write(line)
                    else:
                        mount_for_pid, _, _ = line.split(":")
                        if mount_for_pid == mount_point:
                            cls.shutdown_libfuse_mount(mount_point, os_user, session_dir)
                            mount_point_found = True
                        else:
                            file.write(line)
        except FileNotFoundError:
            log.warning(f"VFS pid file not found at {pid_file_path}")
            return False

        return mount_point_found

    @classmethod
    def get_manifest_path_for_mount(cls, session_dir: Path, mount_point: str) -> Optional[Path]:
        """
        Given a mount_point this searches the pid file for the associated manifest path.

        :param session_dir: tmp directory for session
        :param mount_point: local directory associated with the desired manifest

        :returns: Path to the manifest file for mount if there is one
        """
        try:
            pid_file_path = (session_dir / DEADLINE_VFS_PID_FILE_NAME).resolve()
            with open(pid_file_path, "r") as file:
                for line in file.readlines():
                    line = line.strip()
                    mount_for_pid, _, manifest_path = line.split(":")
                    if mount_for_pid == mount_point:
                        if os.path.exists(manifest_path):
                            return Path(manifest_path)
                        else:
                            log.warning(f"Expected VFS input manifest at {manifest_path}")
                            return None
        except FileNotFoundError:
            log.warning(f"VFS pid file not found at {pid_file_path}")

        log.warning(f"No manifest found for mount {mount_point}")
        return None

    @classmethod
    def is_mount(cls, path) -> bool:
        """
        os.path.ismount returns false for libfuse mounts owned by "other users",
        use findmnt instead
        """
        return subprocess.run(["findmnt", path]).returncode == 0

    @classmethod
    def wait_for_mount(cls, mount_path, session_dir, mount_wait_seconds=60, expected=True) -> bool:
        """
        After we've launched the VFS subprocess we need to wait
        for the OS to validate that the mount is in place before use
        :param mount_path: Path to mount to watch for
        :param session_dir: Session folder associated with mount
        :param mount_wait_seconds: Duration to wait for mount state
        :param expected: Wait for the mount to exist or no longer exist
        """
        log.info(f"Waiting for is_mount at {mount_path} to return {expected}..")
        wait_seconds = mount_wait_seconds
        while wait_seconds >= 0:
            if cls.is_mount(mount_path) == expected:
                log.info(f"is_mount on {mount_path} returns {expected}, returning")
                return True
            wait_seconds -= 1
            if wait_seconds >= 0:
                log.info(f"is_mount on {mount_path} not {expected}, sleeping...")
                time.sleep(1)
        log.info(f"Failed to find is_mount {expected} at {mount_path} after {mount_wait_seconds}")
        cls.print_log_end(session_dir)
        return False

    @classmethod
    def logs_folder_path(cls, session_dir: Path) -> Union[os.PathLike, str]:
        """
        Find the folder we expect VFS logs to be written to
        """
        return session_dir / VFS_LOGS_FOLDER_IN_SESSION

    def get_logs_folder(self) -> Union[os.PathLike, str]:
        """
        Find the folder we expect VFS logs to be written to
        """
        if self._run_path:
            return self.logs_folder_path(Path(self._run_path))
        raise VFSRunPathNotSetError("Attempted to find logs folder without run path")

    @classmethod
    def print_log_end(
        self, session_dir: Path, log_file_name="vfs_log.txt", lines=100, log_level=logging.WARNING
    ):
        """
        Print out the end of our VFS Log.  Reads the full log file into memory.  Our VFS logs are size
        capped so this is not an issue for the intended use case.
        :param session_dir: Session folder for mount
        :param log_file_name: Name of file within the logs folder to read from.  Defaults to vfs_log.txt which
        is our "most recent" log file.
        :param lines: Maximum number of lines from the end of the log to print
        :param log_level: Level to print logging as
        """
        log_file_path = self.logs_folder_path(session_dir) / log_file_name
        log.log(log_level, f"Printing last {lines} lines from {log_file_path}")
        if not os.path.exists(log_file_path):
            log.warning(f"No log file found at {log_file_path}")
            return
        with open(log_file_path, "r") as log_file:
            for this_line in log_file.readlines()[lines * -1 :]:
                log.log(log_level, this_line)

    @classmethod
    def find_vfs_link_dir(cls) -> str:
        """
        Get the path where links to any necessary executables which should be added to the path should live
        :returns: Path to the link folder
        """
        return os.path.join(os.path.dirname(VFSProcessManager.find_vfs()), "..", "link")

    def build_launch_command(self, mount_point: Union[os.PathLike, str]) -> str:
        """
        Build command to pass to Popen to launch VFS
        :param mount_point: directory to mount which must be the first parameter seen by our executable
        :return: command
        """
        executable = VFSProcessManager.find_vfs_launch_script()

        command = (
            f"sudo -E -u {self._os_user}"
            f" {executable} {mount_point} -f --clienttype=deadline"
            f" --bucket={self._asset_bucket}"
            f" --manifest={self._manifest_path}"
            f" --region={self._region}"
            f" -oallow_other"
        )
        if self._cas_prefix is not None:
            command += f" --casprefix={self._cas_prefix}"
        if self._asset_cache_path is not None:
            command += f" --cachedir={self._asset_cache_path}"

        log.info(f"Got launch command {command}")
        return command

    @classmethod
    def find_vfs_launch_script(cls) -> Union[os.PathLike, str]:
        """
        Determine where the VFS launch script lives so we can build the launch command
        :return: Path to VFS launch script
        """
        if VFSProcessManager.launch_script_path is not None:
            log.info(f"Using saved path {VFSProcessManager.launch_script_path} for launch script")
            return VFSProcessManager.launch_script_path

        exe = DEADLINE_VFS_EXECUTABLE
        # for exe in executables:
        log.info(f"Searching for {exe} launch script")
        exe_script = DEADLINE_VFS_EXECUTABLE_SCRIPT
        # Look for env var to construct script path
        if DEADLINE_VFS_ENV_VAR in os.environ:
            log.info(f"{DEADLINE_VFS_ENV_VAR} found in environment")
            environ_check = os.environ[DEADLINE_VFS_ENV_VAR] + exe_script
        else:
            log.warning(f"{DEADLINE_VFS_ENV_VAR} not found in environment")
            environ_check = DEADLINE_VFS_INSTALL_PATH + exe_script
        # Test if script path exists
        if os.path.exists(environ_check):
            log.info(f"Environ check found {exe} launch script at {environ_check}")
            VFSProcessManager.launch_script_path = environ_check
            return environ_check  # type: ignore[return-value]
        else:
            log.error(f"Failed to find {exe} launch script!")

        log.error("Failed to find both executables scripts!")
        raise VFSLaunchScriptMissingError

    @classmethod
    def find_vfs(cls) -> Union[os.PathLike, str]:
        """
        Determine where the VFS executable we'll be launching lives so we can
        find the correct relative paths around it for LD_LIBRARY_PATH and config files
        :return: Path to VFS executable
        """
        if VFSProcessManager.exe_path is not None:
            log.info(f"Using saved path {VFSProcessManager.exe_path}")
            return VFSProcessManager.exe_path

        exe = DEADLINE_VFS_EXECUTABLE
        # Use "which deadline_vfs" by default to find the executable location
        found_path = shutil.which(exe)
        if found_path is None:
            log.info(f"Cwd when finding {exe} is {os.getcwd()}")
            # If VFS executable isn't on the PATH, check if environment variable is set
            if DEADLINE_VFS_ENV_VAR in os.environ:
                log.info(f"{DEADLINE_VFS_ENV_VAR} set to {os.environ[DEADLINE_VFS_ENV_VAR]}")
                environ_check = os.environ[DEADLINE_VFS_ENV_VAR] + f"/bin/{exe}"
            else:
                log.info(f"{DEADLINE_VFS_ENV_VAR} env var not set")
                environ_check = DEADLINE_VFS_INSTALL_PATH + f"/bin/{exe}"
            if os.path.exists(environ_check):
                log.info(f"Environ check found {exe} at {environ_check}")
                found_path = environ_check
            else:
                # Last attempt looks for deadline_vfs in bin
                bin_check = os.path.join(os.getcwd(), f"bin/{exe}")
                if os.path.exists(bin_check):
                    log.info(f"Bin check found VFS at {bin_check}")
                    found_path = bin_check
                else:
                    log.error(f"Failed to find {exe}!")

        # Run final check to see if exe path was found
        if found_path is not None:
            log.info(f"Found {exe} at {found_path}")
            VFSProcessManager.exe_path = found_path
            return found_path  # type: ignore[return-value]

        log.error("Failed to find both executables!")
        raise VFSExecutableMissingError

    @classmethod
    def get_library_path(cls) -> Union[os.PathLike, str]:
        """
        Find our library dependencies which should be at ../lib relative to our executable
        """
        if VFSProcessManager.library_path is None:
            exe_path = VFSProcessManager.find_vfs()
            VFSProcessManager.library_path = os.path.normpath(
                os.path.join(os.path.dirname(exe_path), "../lib")
            )
        log.info(f"Using library path {VFSProcessManager.library_path}")
        return VFSProcessManager.library_path

    def get_file_path(self, relative_file_name: str) -> Union[os.PathLike, str]:
        return os.path.join(self._mount_point, relative_file_name)

    @classmethod
    def create_mount_point(cls, mount_point: Union[os.PathLike, str]) -> None:
        """
        By default fuse won't create our mount folder, create it if it doesn't exist
        """
        if os.path.exists(mount_point) is False:
            log.info(f"Creating mount point at {mount_point}")
            os.makedirs(mount_point, exist_ok=True)
            log.info(f"Modifying permissions of mount point at {mount_point}")
            os.chmod(path=mount_point, mode=0o777)

    @classmethod
    def get_cwd(cls) -> Union[os.PathLike, str]:
        """
        Determine the cwd we should hand to Popen.
        We expect a config/logging.ini file to exist relative to this folder.
        """
        if VFSProcessManager.cwd_path is None:
            exe_path = VFSProcessManager.find_vfs()
            # Use cwd one folder up from bin
            VFSProcessManager.cwd_path = os.path.normpath(
                os.path.join(os.path.dirname(exe_path), "..")
            )
        return VFSProcessManager.cwd_path

    def get_launch_environ(self) -> dict:
        """
        Get the environment variables we'll pass to the launch command.
        :returns: dictionary of default environment variables with VFS changes applied
        """
        my_env = {**self._os_env_vars}
        my_env["PATH"] = f"{VFSProcessManager.find_vfs_link_dir()}{os.pathsep}{os.environ['PATH']}"
        my_env["LD_LIBRARY_PATH"] = VFSProcessManager.get_library_path()  # type: ignore[assignment]
        if os.environ.get(DEADLINE_VFS_CACHE_ENV_VAR) is not None:
            my_env[DEADLINE_VFS_CACHE_ENV_VAR] = os.environ.get(DEADLINE_VFS_CACHE_ENV_VAR)  # type: ignore[assignment]

        return my_env

    def set_manifest_owner(self) -> None:
        """
        Set the manifest path to be owned by _os_user
        """
        log.info(
            f"Attempting to set group ownership on {self._manifest_path} for {self._os_user} to {self._os_group}"
        )
        if not os.path.exists(self._manifest_path):
            log.error(f"Manifest not found at {self._manifest_path}")
            return
        if self._os_group is not None:
            try:
                shutil.chown(self._manifest_path, group=self._os_group)
                os.chmod(self._manifest_path, DEADLINE_MANIFEST_GROUP_READ_PERMS)
            except OSError as e:
                log.error(f"Failed to set ownership with error {e}")
                raise

    def start(self, session_dir: Path) -> None:
        """
        Start our VFS process
        :return: VFS process id
        """
        self._run_path = session_dir
        log.info(f"Using run_path {self._run_path}")
        log.info(f"Using mount_point {self._mount_point}")
        self.set_manifest_owner()
        VFSProcessManager.create_mount_point(self._mount_point)
        start_command = self.build_launch_command(self._mount_point)
        launch_env = self.get_launch_environ()
        log.info(f"Launching VFS with command {start_command}")
        log.info(f"Launching with environment {launch_env}")
        log.info(f"Launching as user {self._os_user}")

        try:

            def read_output_thread(pipe, log):
                # Runs in a thread to redirect VFS output into our log
                try:
                    for line in pipe:
                        log.info(line.decode("utf-8").strip())
                except Exception:
                    log.exception("Error reading VFS output")

            self._vfs_proc = subprocess.Popen(
                args=start_command,
                stdout=subprocess.PIPE,  # Create a new pipe
                stderr=subprocess.STDOUT,  # Merge stderr into the stdout pipe
                cwd=str(self._run_path),
                env=launch_env,
                shell=True,
                executable="/bin/bash",
            )

            self._vfs_thread = threading.Thread(
                target=read_output_thread, args=[self._vfs_proc.stdout, log], daemon=True
            )
            self._vfs_thread.start()

        except Exception as e:
            log.exception(f"Exception during launch with command {start_command} exception {e}")
            raise e
        log.info(f"Launched VFS as pid {self._vfs_proc.pid}")

        is_mounted = VFSProcessManager.wait_for_mount(self.get_mount_point(), session_dir)
        _get_deadline_cloud_library_telemetry_client().record_vfs_mounting(
            successfully_mounted=is_mounted
        )

        if not is_mounted:
            log.error("Failed to mount, shutting down")
            raise VFSFailedToMountError

        try:
            # if the pid file exists, add the new VFS instance and remove any it replaced
            pid_file_path = (session_dir / DEADLINE_VFS_PID_FILE_NAME).resolve()
            with open(pid_file_path, "r") as file:
                lines = file.readlines()
            with open(pid_file_path, "w") as file:
                file.write(f"{self._mount_point}:{self._vfs_proc.pid}:{self._manifest_path}\n")
                for line in lines:
                    line = line.strip()
                    entry_mount_point, entry_pid, entry_manifest_path = line.split(":")
                    if self._mount_point != entry_mount_point:
                        file.write(f"{line}\n")
                    else:
                        log.warning(f"Pid {entry_pid} entry not removed at {entry_mount_point}")
        except FileNotFoundError:
            # if the pid file doesn't exist, this will create it
            with open(pid_file_path, "a") as file:
                file.write(f"{self._mount_point}:{self._vfs_proc.pid}:{self._manifest_path}")

    def get_mount_point(self) -> Union[os.PathLike, str]:
        return self._mount_point



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
The AWS Deadline Cloud CLI interface.
"""

from logging import getLogger
from typing import cast

import click

from . import _groups
from .. import version
from ._main import ContextTrackingGroup, CONTEXT_SETTINGS, deadline as deadline_original

logger = getLogger(__name__)


# Extract all the command names without the trailing "_group" or "_command"
subcommand_names = [v.rsplit("_", 1)[0].replace("_", "-") for v in sorted(_groups.__all__)]
SECTION_LINKS = "\nAvailable sub-commands:\n\n" + "".join(
    f"- [{v}](deadline_{v}.md)\n" for v in subcommand_names
)


# Make a clone of the deadline command, but without the subcommands, for use
# in generating the index.md for the CLI reference documentation. In the CLI reference,
# each subcommand is split out into a separate page, but mkdocs-click always includes
# subcommand docs.
@click.group(
    cls=ContextTrackingGroup,
    context_settings=CONTEXT_SETTINGS,
    help=cast(str, deadline_original.__doc__) + SECTION_LINKS,
)
@click.version_option(version=version, prog_name="deadline")
def deadline():
    pass


deadline.params = deadline_original.params



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

__all__ = ["get_queue_parameter_definitions"]

import yaml

from .. import api
from ._list_apis import _call_paginated_deadline_list_api
from ._session import get_boto3_client
from ..exceptions import DeadlineOperationError
from ..job_bundle.parameters import (
    JobParameter,
    get_ui_control_for_parameter_definition,
    parameter_definition_difference,
    validate_job_parameter,
)
from ..ui._utils import tr


@api.record_function_latency_telemetry_event()
def get_queue_parameter_definitions(
    *, farmId: str, queueId: str, config=None
) -> list[JobParameter]:
    """
    This gets all the queue parameter definitions for the specified [Deadline Cloud queue].

    It does so by getting all the full templates for [queue environments], and then combining
    them equivalently to the Deadline Cloud service logic.

    [Deadline Cloud queue]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/queues.html
    [queue environments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/create-queue-environment.html
    """
    deadline = get_boto3_client("deadline", config=config)
    response = _call_paginated_deadline_list_api(
        deadline.list_queue_environments,
        "environments",
        farmId=farmId,
        queueId=queueId,
    )
    queue_environments = sorted(
        (
            deadline.get_queue_environment(
                farmId=farmId,
                queueId=queueId,
                queueEnvironmentId=queue_env["queueEnvironmentId"],
            )
            for queue_env in response["environments"]
        ),
        key=lambda queue_env: queue_env["priority"],
    )
    queue_environment_templates = [
        yaml.safe_load(queue_env["template"]) for queue_env in queue_environments
    ]

    queue_parameters_definitions: dict[str, JobParameter] = {}
    for template in queue_environment_templates:
        for parameter in template.get("parameterDefinitions", []):
            parameter = validate_job_parameter(parameter, type_required=True, default_required=True)

            # If there is no group label, set it to the name of the Queue Environment
            if not parameter.get("userInterface", {}).get("groupLabel"):
                if "userInterface" not in parameter:
                    parameter["userInterface"] = {
                        "control": get_ui_control_for_parameter_definition(parameter)
                    }
                parameter["userInterface"]["groupLabel"] = tr("Queue Environment: {name}").format(
                    name=template["environment"]["name"]
                )
            existing_parameter = queue_parameters_definitions.get(parameter["name"])
            if existing_parameter:
                differences = parameter_definition_difference(existing_parameter, parameter)
                if differences:
                    raise DeadlineOperationError(
                        f"Job template parameter {parameter['name']} is duplicated across queue environments with mismatched fields:\n"
                        + " ".join(differences)
                    )
            else:
                queue_parameters_definitions[parameter["name"]] = parameter

    return list(queue_parameters_definitions.values())



# coding: utf-8
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""UI Components for the Render Submitter"""

from __future__ import annotations

import logging
import os
import sys
import json
from typing import Any, Dict, Optional, Protocol
import yaml

from qtpy.QtCore import QSize, Qt  # pylint: disable=import-error
from qtpy.QtGui import QKeyEvent  # pylint: disable=import-error
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from .submit_job_progress_dialog import SubmitJobProgressDialog

from ..dataclasses import HostRequirements
from ...dataclasses import SubmitterInfo
from ... import api
from ..deadline_authentication_status import DeadlineAuthenticationStatus
from .._utils import block_signals, tr
from ...config import get_setting, set_setting, config_file
from ...exceptions import UserInitiatedCancel, NonValidInputError
from ...job_bundle import create_job_history_bundle_dir
from ...job_bundle.parameters import JobParameter
from ...job_bundle.submission import AssetReferences
from ..widgets.deadline_authentication_status_widget import DeadlineAuthenticationStatusWidget
from ..widgets.job_attachments_tab import JobAttachmentsWidget
from ..widgets.shared_job_settings_tab import SharedJobSettingsWidget
from ..widgets.host_requirements_tab import HostRequirementsWidget
from . import DeadlineConfigDialog, DeadlineLoginDialog
from ._types import JobBundlePurpose
from ._help_dialog import _HelpDialog

logger = logging.getLogger(__name__)

# initialize early so once the UI opens, things are already initialized
DeadlineAuthenticationStatus.getInstance()


class OnCreateJobBundleCallback(Protocol):
    """This protocol defines the callback for creating a job bundle in the SubmitJobToDeadlineDialog."""

    def __call__(
        self,
        widget: SubmitJobToDeadlineDialog,
        job_bundle_dir: str,
        settings: Any,
        queue_parameters: list[JobParameter],
        asset_references: AssetReferences,
        host_requirements: Optional[Dict[str, Any]] = None,
        *,
        purpose: JobBundlePurpose,
    ) -> Optional[dict[str, Any]]: ...


class SubmitJobToDeadlineDialog(QDialog):
    """
    A widget containing all the standard tabs for submitting an AWS Deadline Cloud job.

    If you're using this dialog within an application and want it to stay in front,
    pass f=Qt.Tool, a flag that tells it to do that.

    Args:
        job_setup_widget_type (QWidget): The type of the widget for the job-specific settings.
        initial_job_settings (dataclass): A dataclass containing the initial job settings
        initial_shared_parameter_values (dict[str, Any]): A dict of parameter values {<name>, <value>, ...}
            to override default queue parameter values from the queue. For example,
            a Rez queue environment may have a default "" for the RezPackages parameter, but a Maya
            submitter would override that default with "maya-2023" or similar.
        auto_detected_attachments (AssetReferences): The job attachments that were automatically detected
            from the input document/scene file or starting job bundle.
        attachments (AssetReferences): The job attachments that have been added to the job by the user.
        on_create_job_bundle_callback (OnCreateJobBundleCallback): A function to call when the dialog
            needs to create a Job Bundle. It is called with arguments:
            (widget, job_bundle_dir, settings, queue_parameters, asset_references, host_requirements, purpose).
            It can return either None or a dict with parameters about the submission. Currently,
            the additional parameters supported are:
            {
                # See documentation for deadline.client.api.create_job_from_job_bundle about these parameters
                "job_parameters": [{"name": "ParameterName", "value": "Parameter Value", ...}],
                "known_asset_paths": ["/path/1", ...],
            }
        parent: parent of the widget
        f: Qt Window Flags
        show_host_requirements_tab: Display the host requirements tab in dialog if set to True. Default
            to False.
        submitter_info (SubmitterInfo): Information related to the submitter window and application it's running in
    """

    def __init__(
        self,
        *,
        job_setup_widget_type: type[QWidget],
        initial_job_settings: Any,
        initial_shared_parameter_values: dict[str, Any],
        auto_detected_attachments: AssetReferences,
        attachments: AssetReferences,
        on_create_job_bundle_callback: OnCreateJobBundleCallback,
        parent: Optional[QWidget] = None,
        f: Qt.WindowFlags = Qt.WindowFlags(),
        show_host_requirements_tab: bool = False,
        host_requirements: Optional[HostRequirements] = None,
        submitter_info: Optional[SubmitterInfo] = None,
        known_asset_paths: Optional[list[str]] = None,
    ):
        # The Qt.Tool flag makes sure our widget stays in front of the main application window
        super().__init__(parent=parent, f=f)

        # Set window title with submitter package info if available
        window_title = tr("Submit to AWS Deadline Cloud")
        if submitter_info:
            # e.g. Deadline Cloud Blender Submitter x.y.z
            formatted_name = f"Deadline Cloud {submitter_info.submitter_name} {tr('Submitter')}"
            if submitter_info.submitter_package_version:
                window_title = f"{formatted_name} {submitter_info.submitter_package_version}"
            else:
                window_title = f"{formatted_name}"
        self.setWindowTitle(window_title)

        self.setMinimumSize(400, 400)

        self.job_settings_type = type(initial_job_settings)
        self.submitter_info = submitter_info or SubmitterInfo(
            submitter_name=self.job_settings_type().submitter_name
        )
        self.on_create_job_bundle_callback = on_create_job_bundle_callback
        self.job_id = None
        self.job_history_bundle_dir: Optional[str] = None
        self.deadline_authentication_status = DeadlineAuthenticationStatus.getInstance()
        self.show_host_requirements_tab = show_host_requirements_tab
        self.known_asset_paths = known_asset_paths or []
        self.should_close = False

        self._build_ui(
            job_setup_widget_type,
            initial_job_settings,
            initial_shared_parameter_values,
            auto_detected_attachments,
            attachments,
            host_requirements,
        )

        self.gui_update_counter: Any = None
        self.refresh_deadline_settings()

    def _submission_succeeded_signal_receiver(self, job_id: str):
        self.job_id = job_id

        set_setting("defaults.job_id", job_id)

    def _close_event_receiver(self):
        if self.submitter_info.submitter_name != "JobBundle" and self.job_id:
            self.close()

    def sizeHint(self):
        return QSize(540, 700)

    def refresh(
        self,
        *,
        job_settings: Optional[Any] = None,
        auto_detected_attachments: Optional[AssetReferences] = None,
        attachments: Optional[AssetReferences] = None,
        load_new_bundle: bool = False,
    ):
        # Refresh the UI components
        self.refresh_deadline_settings()
        if (auto_detected_attachments is not None) or (attachments is not None):
            self.job_attachments.refresh_ui(auto_detected_attachments, attachments)

        if job_settings is not None:
            self.job_settings_type = type(job_settings)
            # Refresh shared job settings
            self.shared_job_settings.refresh_ui(job_settings, load_new_bundle)
            # Refresh job specific settings
            if hasattr(self.job_settings, "refresh_ui"):
                self.job_settings.refresh_ui(job_settings)

    def _build_ui(
        self,
        job_setup_widget_type,
        initial_job_settings,
        initial_shared_parameter_values,
        auto_detected_attachments: AssetReferences,
        attachments: AssetReferences,
        host_requirements: Optional[HostRequirements],
    ):
        self.lyt = QVBoxLayout(self)
        self.lyt.setContentsMargins(5, 5, 5, 5)

        man_layout = QFormLayout()
        self.lyt.addLayout(man_layout)
        self.tabs = QTabWidget()
        self.lyt.addWidget(self.tabs)

        self._build_shared_job_settings_tab(initial_job_settings, initial_shared_parameter_values)
        self._build_job_settings_tab(job_setup_widget_type, initial_job_settings)
        self._build_job_attachments_tab(auto_detected_attachments, attachments)

        # Show host requirements only if requested by the constructor
        if self.show_host_requirements_tab:
            self._build_host_requirements_tab(host_requirements)

        self.auth_status_box = DeadlineAuthenticationStatusWidget(self)
        self.auth_status_box.switch_profile_clicked.connect(self.on_switch_profile_clicked)
        self.auth_status_box.logout_clicked.connect(self.on_logout)
        self.auth_status_box.login_clicked.connect(self.on_login)
        self.lyt.addWidget(self.auth_status_box)
        self.deadline_authentication_status.api_availability_changed.connect(
            self.refresh_deadline_settings
        )

        # Refresh the submit button enable state once queue parameter status changes
        self.shared_job_settings.valid_parameters.connect(self._set_submit_button_state)

        self.button_box = QDialogButtonBox(Qt.Horizontal)
        self.settings_button = QPushButton(tr("Settings..."))
        self.settings_button.clicked.connect(self.on_settings_button_clicked)
        self.button_box.addButton(self.settings_button, QDialogButtonBox.ResetRole)
        self.help_button = QPushButton(tr("Help"))
        self.help_button.clicked.connect(self._on_help_button_clicked)
        self.button_box.addButton(self.help_button, QDialogButtonBox.HelpRole)
        self.submit_button = QPushButton(tr("Submit"))
        self.submit_button.clicked.connect(self.on_submit)
        self.button_box.addButton(self.submit_button, QDialogButtonBox.AcceptRole)
        if hasattr(initial_job_settings, "browse_enabled") and initial_job_settings.browse_enabled:
            self.load_bundle_button = QPushButton(tr("Load Bundle"))
            self.load_bundle_button.clicked.connect(self._on_load_bundle)
            self.button_box.addButton(self.load_bundle_button, QDialogButtonBox.AcceptRole)
        self.export_bundle_button = QPushButton(tr("Export bundle"))
        self.export_bundle_button.clicked.connect(self.on_export_bundle)
        self.button_box.addButton(self.export_bundle_button, QDialogButtonBox.AcceptRole)

        self.lyt.addWidget(self.button_box)

    def _set_submit_button_state(self):
        # Enable/disable the Submit button based on whether the
        # AWS Deadline Cloud API is accessible and the farm+queue are configured.
        api_available = self.deadline_authentication_status.api_availability is True
        farm_configured = get_setting("defaults.farm_id") != ""
        queue_configured = get_setting("defaults.queue_id") != ""
        queue_valid = self.shared_job_settings.is_queue_valid()

        enable = api_available and farm_configured and queue_configured and queue_valid

        self.submit_button.setEnabled(enable)

        if not enable:
            issues = []
            if not api_available:
                issues.append(
                    tr(
                        "AWS Deadline Cloud API is not accessible. Check your authentication status."
                    )
                )
            if not farm_configured:
                issues.append(
                    tr("No farm is configured. Click Settings to select a farm for job submission.")
                )
            if not queue_configured:
                issues.append(
                    tr("No queue is configured. Click Settings to select a queue within your farm.")
                )
            if farm_configured and queue_configured and not queue_valid:
                issues.append(
                    tr("Queue parameters are not valid. Check Shared job settings tab for details.")
                )

            self.submit_button.setToolTip(
                tr("Cannot submit job:\n\n\u2022 {issues}").format(
                    issues="\n\n\u2022 ".join(issues)
                )
            )
        else:
            self.submit_button.setToolTip("")

    def refresh_deadline_settings(self):
        self._set_submit_button_state()

        self.shared_job_settings.deadline_cloud_settings_box.refresh_setting_controls(
            self.deadline_authentication_status.api_availability is True
        )
        # If necessary, this reloads the queue parameters
        self.shared_job_settings.refresh_queue_parameters()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Override to capture any enter/return key presses so that the Submit
        button isn't "pressed" when the enter/return key is.
        """
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            return
        super().keyPressEvent(event)

    def _build_shared_job_settings_tab(self, initial_job_settings, initial_shared_parameter_values):
        self.shared_job_settings_tab = QScrollArea()
        self.tabs.addTab(self.shared_job_settings_tab, tr("Shared job settings"))
        self.shared_job_settings = SharedJobSettingsWidget(
            initial_settings=initial_job_settings,
            initial_shared_parameter_values=initial_shared_parameter_values,
            parent=self,
        )
        self.shared_job_settings.parameter_changed.connect(self.on_shared_job_parameter_changed)
        self.shared_job_settings_tab.setWidget(self.shared_job_settings)
        self.shared_job_settings_tab.setWidgetResizable(True)
        self.shared_job_settings.parameter_changed.connect(self.on_shared_job_parameter_changed)

    def _build_job_settings_tab(self, job_setup_widget_type, initial_job_settings):
        self.job_settings_tab = QScrollArea()
        self.tabs.addTab(self.job_settings_tab, tr("Job-specific settings"))
        self.job_settings_tab.setWidgetResizable(True)

        self.job_settings = job_setup_widget_type(
            initial_settings=initial_job_settings, parent=self
        )
        self.job_settings_tab.setWidget(self.job_settings)
        if hasattr(self.job_settings, "parameter_changed"):
            self.job_settings.parameter_changed.connect(self.on_job_template_parameter_changed)

    def _build_job_attachments_tab(
        self, auto_detected_attachments: AssetReferences, attachments: AssetReferences
    ):
        self.job_attachments_tab = QScrollArea()
        self.tabs.addTab(self.job_attachments_tab, tr("Job attachments"))
        self.job_attachments = JobAttachmentsWidget(
            auto_detected_attachments, attachments, parent=self
        )
        self.job_attachments_tab.setWidget(self.job_attachments)
        self.job_attachments_tab.setWidgetResizable(True)

    def _build_host_requirements_tab(self, host_requirements: Optional[HostRequirements]):
        self.host_requirements = HostRequirementsWidget()
        self.host_requirements_tab = QScrollArea()
        self.tabs.addTab(self.host_requirements_tab, tr("Host requirements"))
        self.host_requirements_tab.setWidget(self.host_requirements)
        self.host_requirements_tab.setWidgetResizable(True)
        if host_requirements:
            self.host_requirements.set_requirements(host_requirements)

    def on_shared_job_parameter_changed(self, parameter: dict[str, Any]):
        """
        Handles an edit to a shared job parameter, for example one of the
        queue parameters.

        When a queue parameter and a job template parameter have
        the same name, we update between them to keep them consistent.
        """
        try:
            if hasattr(self.job_settings, "set_parameter_value"):
                with block_signals(self.job_settings):
                    self.job_settings.set_parameter_value(parameter)
        except KeyError:
            # If there is no corresponding job template parameter,
            # just ignore it.
            pass

    def on_job_template_parameter_changed(self, parameter: dict[str, Any]):
        """
        Handles an edit to a job template parameter.

        When a queue parameter and a job template parameter have
        the same name, we update between them to keep them consistent.
        """
        try:
            with block_signals(self.shared_job_settings):
                self.shared_job_settings.set_parameter_value(parameter)
        except KeyError:
            # If there is no corresponding queue parameter,
            # just ignore it.
            pass

    def on_login(self):
        DeadlineLoginDialog.login(parent=self)
        self.refresh_deadline_settings()
        # This widget watches the auth files, but that does
        # not always catch a change so force a refresh here.
        self.deadline_authentication_status.refresh_status()

    def on_logout(self):
        api.logout()
        self.refresh_deadline_settings()
        # This widget watches the auth files, but that does
        # not always catch a change so force a refresh here.
        self.deadline_authentication_status.refresh_status()

    def on_switch_profile_clicked(self):
        if DeadlineConfigDialog.configure_settings(parent=self, set_profile_focus=True):
            self.refresh_deadline_settings()

    def on_settings_button_clicked(self):
        if DeadlineConfigDialog.configure_settings(parent=self):
            self.refresh_deadline_settings()

    def _on_help_button_clicked(self):
        """Show the Help dialog with submitter information."""
        try:
            dialog = _HelpDialog(self.submitter_info, parent=self)
            dialog.exec_()
        except Exception as e:
            logger.error(f"Failed to create HelpDialog: {e}")
            QMessageBox.critical(
                self,
                "Error",
                f"Failed to display Help dialog: {str(e)}",
            )

    def _on_load_bundle(self):
        """Delegates to the job_settings widget's on_load_bundle method."""
        if hasattr(self.job_settings, "on_load_bundle"):
            self.job_settings.on_load_bundle()

    def on_export_bundle(self):
        """
        Exports a Job Bundle, but does not submit the job.
        """
        # Retrieve all the settings into the dataclass
        settings = self.job_settings_type()
        self.shared_job_settings.update_settings(settings)
        self.job_settings.update_settings(settings)

        queue_parameters = self.shared_job_settings.get_parameters()

        asset_references = self.job_attachments.get_asset_references()

        # Save the bundle
        try:
            self.job_history_bundle_dir = create_job_history_bundle_dir(
                self.submitter_info.submitter_name, settings.name
            )

            if self.show_host_requirements_tab:
                host_requirements = self.host_requirements.get_requirements()
                parameters_from_callback = self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    host_requirements,
                    purpose=JobBundlePurpose.EXPORT,
                )
            else:
                # Maintaining backward compatibility for submitters that do not support host_requirements yet
                parameters_from_callback = self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    purpose=JobBundlePurpose.EXPORT,
                )
            if parameters_from_callback is None:
                parameters_from_callback = {}

            # If the callback returned job parameters, update them in the job bundle as well so that
            # submission from the job history dir is equivalent.
            job_parameters = parameters_from_callback.get("job_parameters", [])
            if job_parameters:
                self.save_job_parameters_to_job_bundle(self.job_history_bundle_dir, job_parameters)

            logger.info(f"Saved the submission as a job bundle: {self.job_history_bundle_dir}")
            if sys.platform == "win32":
                # Open the directory in the OS's file explorer
                os.startfile(self.job_history_bundle_dir)
            QMessageBox.information(
                self,
                tr("{submitter} job submission").format(
                    submitter=self.submitter_info.submitter_name
                ),
                tr("Saved the submission as a job bundle:\n{path}").format(
                    path=self.job_history_bundle_dir
                ),
            )
            # Close the submitter window to signal the submission is done
            self.close()

        except NonValidInputError as nvie:
            QMessageBox.critical(self, tr("Non valid inputs detected"), str(nvie))

        except Exception as exc:
            logger.exception("Error saving bundle")
            message = str(exc)
            QMessageBox.critical(
                self,
                tr("{submitter} job submission").format(
                    submitter=self.submitter_info.submitter_name
                ),
                message,
            )  # type: ignore[call-arg]

    def save_job_parameters_to_job_bundle(
        self, job_bundle_dir: str, job_parameters: list[JobParameter]
    ):
        """
        Saves the job parameters to the job bundle. If the job bundle already has a parameter_values file,
        it updates it. Otherwise it creates it.
        """
        job_parameters_dict = {param["name"]: param for param in job_parameters}

        job_parameters_file = os.path.join(job_bundle_dir, "parameter_values.yaml")
        if os.path.exists(job_parameters_file):
            with open(job_parameters_file, "r", encoding="utf8") as f:
                existing_job_parameters = yaml.safe_load(f).get("parameterValues", [])
        else:
            job_parameters_file = os.path.join(job_bundle_dir, "parameter_values.json")
            if os.path.exists(job_parameters_file):
                with open(job_parameters_file, "r", encoding="utf8") as f:
                    existing_job_parameters = json.load(f).get("parameterValues", [])
            else:
                existing_job_parameters = []

        # Overwrite any existing values, and add new values at the end
        combined_job_parameters = []
        for param in existing_job_parameters:
            combined_job_parameters.append(job_parameters_dict.pop(param["name"], param))
        combined_job_parameters.extend(job_parameters_dict.values())

        with open(job_parameters_file, "w", encoding="utf8") as f:
            json.dump({"parameterValues": combined_job_parameters}, f, indent=1)

    def on_submit(self):
        """
        Perform a submission when the submit button is pressed
        """
        # Retrieve all the settings into the dataclass
        settings = self.job_settings_type()
        self.shared_job_settings.update_settings(settings)
        self.job_settings.update_settings(settings)

        queue_parameters = self.shared_job_settings.get_parameters()

        asset_references = self.job_attachments.get_asset_references()

        job_progress_dialog = SubmitJobProgressDialog(parent=self)
        job_progress_dialog.submission_thread_succeeded.connect(
            self._submission_succeeded_signal_receiver
        )
        job_progress_dialog.progress_window_closed.connect(self._close_event_receiver)
        job_progress_dialog.setModal(True)
        job_progress_dialog.show()
        QApplication.instance().processEvents()  # type: ignore[union-attr]

        # Submit the job
        try:
            self.job_history_bundle_dir = create_job_history_bundle_dir(
                self.submitter_info.submitter_name, settings.name
            )

            if self.show_host_requirements_tab:
                requirements = self.host_requirements.get_requirements()
                parameters_from_callback = self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    requirements,
                    purpose=JobBundlePurpose.SUBMISSION,
                )
            else:
                # Maintaining backward compatibility for submitters that do not support host_requirements yet
                parameters_from_callback = self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    purpose=JobBundlePurpose.SUBMISSION,
                )
            if parameters_from_callback is None:
                parameters_from_callback = {}

            # If the callback returned job parameters, update them in the job bundle as well so that
            # submission from the job history dir is equivalent.
            job_parameters = parameters_from_callback.get("job_parameters", [])
            if job_parameters:
                self.save_job_parameters_to_job_bundle(self.job_history_bundle_dir, job_parameters)

            job_progress_dialog.start_job_submission(
                job_bundle_dir=self.job_history_bundle_dir,
                submitter_name=self.submitter_info.submitter_name,
                config=config_file.read_config(),
                require_paths_exist=self.job_attachments.get_require_paths_exist(),
                job_parameters=job_parameters,
                known_asset_paths=self.known_asset_paths
                + parameters_from_callback.get("known_asset_paths", []),
            )

        except UserInitiatedCancel as uic:
            logger.info("Canceling submission.")
            QMessageBox.information(
                self,
                tr("{submitter} job submission").format(
                    submitter=self.submitter_info.submitter_name
                ),
                str(uic),
            )
            job_progress_dialog.close()
        except NonValidInputError as nvie:
            QMessageBox.critical(self, tr("Non valid inputs detected"), str(nvie))
            job_progress_dialog.close()
        except Exception as exc:
            logger.exception("error submitting job")
            api.get_deadline_cloud_library_telemetry_client().record_error(
                event_details={"exception_scope": "on_submit"},
                exception_type=str(type(exc)),
                from_gui=True,
            )
            QMessageBox.critical(
                self,
                tr("{submitter} job submission").format(
                    submitter=self.submitter_info.submitter_name
                ),
                str(exc),
            )  # type: ignore[call-arg]
            job_progress_dialog.close()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Helper functions for suggesting available resources when CLI commands fail
due to incorrect resource IDs.
"""

from __future__ import annotations

from configparser import ConfigParser
from typing import Optional

from botocore.exceptions import ClientError

from .. import api


def _format_resource_suggestions(
    items: list,
    id_field: str,
    name_field: str,
    header: str,
    max_items: int = 10,
) -> list[str]:
    """Format a list of resources as suggestion lines."""
    if not items:
        return []
    lines = [f"\n{header}"]
    for item in items[:max_items]:
        lines.append(f"  {item[id_field]}  {item.get(name_field, '')}")
    if len(items) > max_items:
        lines.append(f"  ... and {len(items) - max_items} more")
    return lines


def _try_suggestion_chain(
    fetchers: list[tuple],
    config: Optional[ConfigParser],
) -> tuple[list[str], bool]:
    """
    Try a chain of fetchers until one succeeds.
    Each fetcher is a tuple of (fetch_func, response_key, id_field, name_field, header).
    Returns (suggestion_lines, all_failed).
    """
    for fetch_func, response_key, id_field, name_field, header in fetchers:
        try:
            response = fetch_func()
            items = response.get(response_key, [])
            suggestions = _format_resource_suggestions(items, id_field, name_field, header)
            if suggestions:
                return suggestions, False
        except ClientError:
            continue
    return [], True


# Mapping of operation names to their suggestion configuration.
_OPERATION_GROUPS: dict[str, dict[str, tuple]] = {
    "queue": {
        "operations": ("GetQueue", "ListQueues", "ListQueueEnvironments"),
        "requires": ("farm_id",),
    },
    "farm": {
        "operations": ("GetFarm", "ListFarms"),
        "requires": (),
    },
    "fleet": {
        "operations": ("GetFleet", "ListFleets"),
        "requires": ("farm_id",),
    },
    "worker": {
        "operations": ("GetWorker", "SearchWorkers"),
        "requires": ("farm_id", "fleet_id"),
    },
    "job": {
        "operations": ("GetJob", "ListJobs", "SearchJobs"),
        "requires": ("farm_id", "queue_id"),
    },
    "storage_profile": {
        "operations": ("GetStorageProfileForQueue", "ListStorageProfilesForQueue"),
        "requires": ("farm_id", "queue_id"),
    },
}


def _suggest_resources_on_client_error(
    exc: ClientError,
    farm_id: Optional[str] = None,
    queue_id: Optional[str] = None,
    fleet_id: Optional[str] = None,
    worker_id: Optional[str] = None,
    config: Optional[ConfigParser] = None,
) -> str:
    """
    When a ClientError occurs (e.g., AccessDeniedException), attempt to list
    available resources to help users identify typos in resource IDs.

    Returns a suggestion string to append to the error message, or empty string
    if no suggestions could be generated.
    """
    error_code = exc.response.get("Error", {}).get("Code", "")

    # Only handle access/not-found type errors that indicate a wrong resource ID
    if error_code not in (
        "AccessDeniedException",
        "ResourceNotFoundException",
        "ValidationException",
    ):
        return ""

    operation_name = getattr(exc, "operation_name", "") or ""

    # Build fetcher functions that capture the config and resource IDs
    def list_farms():
        return api.list_farms(config=config)

    def list_queues():
        return api.list_queues(farmId=farm_id, config=config)

    def list_fleets():
        return api.list_fleets(farmId=farm_id, config=config)

    def list_jobs():
        return api.list_jobs(farmId=farm_id, queueId=queue_id, config=config)

    def list_storage_profiles():
        return api.list_storage_profiles_for_queue(farmId=farm_id, queueId=queue_id, config=config)

    def search_workers():
        deadline = api.get_boto3_client("deadline", config=config)
        search_kwargs: dict = {"farmId": farm_id, "fleetIds": [fleet_id], "itemOffset": 0}
        if worker_id:
            search_kwargs["filterExpressions"] = {
                "filters": [{"searchTermFilter": {"searchTerm": worker_id}}],
                "operator": "AND",
            }
        return deadline.search_workers(**search_kwargs)

    # Define suggestion chains for each operation group
    suggestions: list[str] = []
    list_failed = False

    if operation_name in _OPERATION_GROUPS["queue"]["operations"] and farm_id:
        suggestions, list_failed = _try_suggestion_chain(
            [
                (
                    list_queues,
                    "queues",
                    "queueId",
                    "displayName",
                    f"Available queues in farm {farm_id}:",
                ),
                (
                    list_farms,
                    "farms",
                    "farmId",
                    "displayName",
                    f"Farm {farm_id} may be incorrect. Available farms:",
                ),
            ],
            config,
        )

    elif operation_name in _OPERATION_GROUPS["farm"]["operations"]:
        suggestions, list_failed = _try_suggestion_chain(
            [(list_farms, "farms", "farmId", "displayName", "Available farms:")],
            config,
        )

    elif operation_name in _OPERATION_GROUPS["fleet"]["operations"] and farm_id:
        suggestions, list_failed = _try_suggestion_chain(
            [
                (
                    list_fleets,
                    "fleets",
                    "fleetId",
                    "displayName",
                    f"Available fleets in farm {farm_id}:",
                ),
                (
                    list_farms,
                    "farms",
                    "farmId",
                    "displayName",
                    f"Farm {farm_id} may be incorrect. Available farms:",
                ),
            ],
            config,
        )

    elif operation_name in _OPERATION_GROUPS["worker"]["operations"] and farm_id and fleet_id:
        # Workers use search_workers with special handling for totalResults
        try:
            response = search_workers()
            workers = response.get("workers", [])
            if workers:
                suggestions = [f"\nAvailable workers in fleet {fleet_id}:"]
                for w in workers[:10]:
                    suggestions.append(f"  {w['workerId']}  {w.get('status', '')}")
                total = response.get("totalResults", len(workers))
                if total > 10:
                    suggestions.append(f"  ... and {total - 10} more")
        except ClientError:
            suggestions, list_failed = _try_suggestion_chain(
                [
                    (
                        list_fleets,
                        "fleets",
                        "fleetId",
                        "displayName",
                        f"Fleet {fleet_id} may be incorrect. Available fleets:",
                    )
                ],
                config,
            )

    elif operation_name in _OPERATION_GROUPS["job"]["operations"] and farm_id and queue_id:
        suggestions, list_failed = _try_suggestion_chain(
            [
                (list_jobs, "jobs", "jobId", "name", f"Recent jobs in queue {queue_id}:"),
                (
                    list_queues,
                    "queues",
                    "queueId",
                    "displayName",
                    f"Queue {queue_id} may be incorrect. Available queues:",
                ),
                (
                    list_farms,
                    "farms",
                    "farmId",
                    "displayName",
                    f"Farm {farm_id} may be incorrect. Available farms:",
                ),
            ],
            config,
        )

    elif (
        operation_name in _OPERATION_GROUPS["storage_profile"]["operations"]
        and farm_id
        and queue_id
    ):
        suggestions, list_failed = _try_suggestion_chain(
            [
                (
                    list_storage_profiles,
                    "storageProfiles",
                    "storageProfileId",
                    "displayName",
                    f"Available storage profiles for queue {queue_id}:",
                )
            ],
            config,
        )

    # If we couldn't list resources, add a hint about permissions
    if list_failed and not suggestions:
        suggestions = [
            "\nCould not list available resources to suggest alternatives.",
            "This may indicate your IAM policy is missing List permissions.",
        ]

    return "\n".join(suggestions)
